##
# Utility Functions to support re-use in python scripts.
# Includes functions for running external commands, etc
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Module containing utility functions to support re-use in python scripts."""

import datetime
import importlib
import inspect
import logging
import os
import platform
import re
import shutil
import stat
import subprocess
import sys
import threading
import time
from collections import namedtuple
from io import BytesIO, StringIO
from typing import Any, Callable, Optional
from warnings import warn


# https://stackoverflow.com/questions/2829329/catch-a-threads-exception-in-the-caller-thread-in-python
# TODO make this a private class.
class PropagatingThread(threading.Thread):  # noqa
    """Class to support running commands from the shell in a python environment.

    Don't use directly.

    PropagatingThread copied from sample here:
    """

    def run(self):  # noqa
        self.exc = None
        try:
            if hasattr(self, "_Thread__target"):
                # Thread uses name mangling prior to Python 3.
                self.ret = self._Thread__target(*self._Thread__args, **self._Thread__kwargs)
            else:
                self.ret = self._target(*self._args, **self._kwargs)
        except BaseException as e:
            self.exc = e

    def join(self, timeout=None):  # noqa
        super(PropagatingThread, self).join()
        if self.exc:
            raise self.exc
        return self.ret


# http://stackoverflow.com/questions/19423008/logged-subprocess-communicate
# TODO make this a private function
def reader(filepath, outstream, stream, logging_level=logging.INFO, encodingErrors="strict"):  # noqa
    """Helper functions for running commands from the shell in python environment.

    Don't use directly

    process output stream and write to log.
    part of the threading pattern.
    """
    f = None
    # open file if caller provided path
    if filepath:
        f = open(filepath, "w")

    while True:
        s = stream.readline().decode(sys.stdout.encoding or "utf-8", errors=encodingErrors)
        if not s:
            break
        if f is not None:
            # write to file if caller provided file
            f.write(s)
        if outstream is not None:
            # write to stream object if caller provided object
            outstream.write(s)
        logging.log(logging_level, s.rstrip())
    stream.close()
    if f is not None:
        f.close()


def GetHostInfo() -> namedtuple:
    """Returns a namedtuple containing information about host machine.

    Returns:
        (namedTuple[Host(os, arch, bit)]): Host(os=OS Type, arch=System Architecture, bit=Highest Order Bit)
    """
    Host = namedtuple("Host", "os arch bit")
    host_info = platform.uname()
    os = host_info.system
    processor_info = host_info.machine

    arch = None
    bit = None

    # For backwards compatibility, platform.uname() returns Darwin for macs, which is the kernal rather then the OS.
    # Due to this, we manually fix it.
    if os == "Darwin":
        os = "MacOs"

    if ("x86" in processor_info.lower()) or ("AMD" in processor_info.upper()) or ("INTEL" in processor_info.upper()):
        arch = "x86"
    elif ("ARM" in processor_info.upper()) or ("AARCH" in processor_info.upper()):
        arch = "ARM"

    if "32" in processor_info:
        bit = "32"
    elif "64" in processor_info:
        bit = "64"

    if (arch is None) or (bit is None):
        raise EnvironmentError("Host info could not be parsed: {0}".format(str(host_info)))

    return Host(os=os, arch=arch, bit=bit)


def timing(f: Callable) -> Callable:
    """This is a mixing to do timing on a function.

    Example:
        ```
            @timing
            def function_i_want_to_time():
        ```
    """

    def wrap(*args: Any) -> int:  #
        time1 = time.time()
        ret = f(*args)
        time2 = time.time()
        logging.debug("{:s} function took {:.3f} ms".format(f.__name__, (time2 - time1) * 1000.0))
        return ret

    return wrap


def RunCmd(
    cmd: str,
    parameters: str,
    capture: bool = True,
    workingdir: str = None,
    outfile: Optional[str] = None,
    outstream: Optional[str] = None,
    environ: Optional[dict] = None,
    logging_level: int = logging.INFO,
    raise_exception_on_nonzero: bool = False,
    encodingErrors: str = "strict",
    close_fds: bool = True,
) -> int:
    """Run a shell command and print the output to the log file.

    This is the public function that should be used to run commands from the shell in python environment

    Attributes:
        cmd: command being run, either quoted or not quoted
        parameters: parameters string taken as is
        capture (obj): boolean to determine if caller wants the output captured
            in any format.
        workingdir: path to set to the working directory before running
            the command.
        outfile: capture output to file of given path.
        outstream: capture output to a stream.
        environ: shell environment variables dictionary that replaces the
            one inherited from the current process.
        logging_level: log level to log output at.  Default is INFO
        raise_exception_on_nonzero: Setting to true causes exception to
            be raised if the cmd return code is not zero.
        encodingErrors: may be given to set the desired error handling
            for encoding errors decoding cmd output. Default is 'strict'.
        close_fds: If True, file descriptors are closed before the
            command is run. Default is True.

    Returns:
        (int): returncode of called cmd
    """
    cmd = cmd.strip("\"'")
    if " " in cmd:
        cmd = '"' + cmd + '"'
    if parameters is not None:
        parameters = parameters.strip()
        cmd += " " + parameters
    starttime = datetime.datetime.now()
    logging.log(logging_level, "Cmd to run is: " + cmd)
    logging.log(logging_level, "------------------------------------------------")
    logging.log(logging_level, "--------------Cmd Output Starting---------------")
    logging.log(logging_level, "------------------------------------------------")
    c = subprocess.Popen(
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        cwd=workingdir,
        shell=True,
        env=environ,
        close_fds=close_fds,
    )
    if capture:
        thread = PropagatingThread(target=reader, args=(outfile, outstream, c.stdout, logging_level, encodingErrors))
        thread.start()
        c.wait()
        thread.join()
    else:
        c.wait()

    endtime = datetime.datetime.now()
    delta = endtime - starttime
    endtime_str = "{0[0]:02}:{0[1]:02}".format(divmod(delta.seconds, 60))
    returncode_str = "{0:#010x}".format(c.returncode)
    logging.log(logging_level, "------------------------------------------------")
    logging.log(logging_level, "--------------Cmd Output Finished---------------")
    logging.log(logging_level, "--------- Running Time (mm:ss): " + endtime_str + " ----------")
    logging.log(logging_level, "----------- Return Code: " + returncode_str + " ------------")
    logging.log(logging_level, "------------------------------------------------")

    if raise_exception_on_nonzero and c.returncode != 0:
        raise Exception("{0} failed with Return Code: {1}".format(cmd, returncode_str))
    return c.returncode


def RunPythonScript(
    pythonfile: str,
    params: str,
    capture: bool = True,
    workingdir: Optional[str] = None,
    outfile: Optional[str] = None,
    outstream: Optional[str] = None,
    environ: Optional[dict] = None,
    logging_level: int = logging.INFO,
    raise_exception_on_nonzero: bool = False,
) -> int:
    """Run a python script and print the output to the log file.

    This is the public function that should be used to execute python scripts from the shell in python environment.
    The python script will be located using the path as if it was an executable.

    Args:
        pythonfile: python file to execute
        params: parameters string taken as is
        capture: if caller wants the output captured in any format.
        workingdir: path to set to the working directory before running the command.
        outfile: capture output to file of given path.
        outstream: capture output to a stream.
        environ: shell environment variables dictionary that replaces the one inherited from the current process.
        logging_level: log level to log output at.  Default is INFO
        raise_exception_on_nonzero: causes exception to be raised if the cmd return code is not zero.

    Returns:
        (int): returncode of called cmd
    """
    # locate python file on path
    pythonfile.strip("\"'")
    if " " in pythonfile:
        pythonfile = '"' + pythonfile + '"'
    params.strip()
    logging.debug("RunPythonScript: {0} {1}".format(pythonfile, params))
    if os.path.isabs(pythonfile):
        logging.debug("Python Script was given as absolute path: %s" % pythonfile)
    elif os.path.isfile(os.path.join(os.getcwd(), pythonfile)):
        pythonfile = os.path.join(os.getcwd(), pythonfile)
        logging.debug("Python Script was given as relative path: %s" % pythonfile)
    else:
        # loop thru path environment variable
        for a in os.getenv("PATH").split(os.pathsep):
            a = os.path.normpath(a)
            if os.path.isfile(os.path.join(a, pythonfile)):
                pythonfile = os.path.join(a, pythonfile)
                logging.debug("Python Script was found on the path: %s" % pythonfile)
                break
    params = pythonfile + " " + params
    return RunCmd(
        sys.executable,
        params,
        capture=capture,
        workingdir=workingdir,
        outfile=outfile,
        outstream=outstream,
        environ=environ,
        logging_level=logging_level,
        raise_exception_on_nonzero=raise_exception_on_nonzero,
    )


def DetachedSignWithSignTool(
    SignToolPath: str,
    ToSignFilePath: str,
    SignatureOutputFile: str,
    PfxFilePath: str,
    PfxPass: Optional[str] = None,
    Oid: str = "1.2.840.113549.1.7.2",
    Eku: Optional[str] = None,
) -> int:
    """Locally Sign input file using Windows SDK signtool.

    This will use a local Pfx file.
    WARNING: This should not be used for production signing as that process should follow stronger
        security practices (HSM / smart cards / etc)

    Signing is in format specified by UEFI authentacted variables
    """
    # check signtool path
    if not os.path.exists(SignToolPath):
        logging.error("Path to signtool invalid.  %s" % SignToolPath)
        return -1

    # Adjust for spaces in the path (when calling the command).
    if " " in SignToolPath:
        SignToolPath = '"' + SignToolPath + '"'

    OutputDir = os.path.dirname(SignatureOutputFile)
    # Signtool docs https://docs.microsoft.com/en-us/dotnet/framework/tools/signtool-exe
    # Signtool parameters from
    #   https://docs.microsoft.com/en-us/windows-hardware/manufacture/desktop/secure-boot-key-generation-and-signing-using-hsm--example  # noqa: E501
    # Search for "Secure Boot Key Generation and Signing Using HSM"
    params = (
        "sign /fd sha256 /p7ce DetachedSignedData /p7co " + Oid + ' /p7 "' + OutputDir + '" /f "' + PfxFilePath + '"'
    )
    if Eku is not None:
        params += " /u " + Eku
    if PfxPass is not None:
        # add password if set
        params = params + " /p " + PfxPass
    params = params + ' /debug /v "' + ToSignFilePath + '" '
    ret = RunCmd(SignToolPath, params)
    if ret != 0:
        logging.error("Signtool failed %d" % ret)
        return ret
    signedfile = os.path.join(OutputDir, os.path.basename(ToSignFilePath) + ".p7")
    if not os.path.isfile(signedfile):
        raise Exception("Output file doesn't exist %s" % signedfile)

    shutil.move(signedfile, SignatureOutputFile)
    return ret


def CatalogSignWithSignTool(
    SignToolPath: str, ToSignFilePath: str, PfxFilePath: str, PfxPass: Optional[str] = None
) -> int:
    """Locally sign input file using Windows SDK signtool.

    This will use a local Pfx file.

    WARNING: This should not be used for production signing as that process should follow
        stronger security practices (HSM / smart cards / etc)

    Signing is catalog format which is an attached signature
    """
    # check signtool path
    if not os.path.exists(SignToolPath):
        logging.error("Path to signtool invalid.  %s" % SignToolPath)
        return -1

    # Adjust for spaces in the path (when calling the command).
    if " " in SignToolPath:
        SignToolPath = '"' + SignToolPath + '"'

    OutputDir = os.path.dirname(ToSignFilePath)
    # Signtool docs https://docs.microsoft.com/en-us/dotnet/framework/tools/signtool-exe
    # todo: link to catalog signing documentation
    params = "sign /a /fd SHA256 /f " + PfxFilePath
    if PfxPass is not None:
        # add password if set
        params = params + " /p " + PfxPass
    params = params + ' /debug /v "' + ToSignFilePath + '" '
    ret = RunCmd(SignToolPath, params, workingdir=OutputDir)
    if ret != 0:
        logging.error("Signtool failed %d" % ret)
    return ret


# Simplified Comparison Function borrowed from StackOverflow...
# https://stackoverflow.com/questions/1714027/version-number-comparison
# With Python 3.0 help from:
# https://docs.python.org/3.0/whatsnew/3.0.html#ordering-comparisons
def version_compare(version1: str, version2: str) -> bool:
    """Compare two versions."""

    def normalize(v: str) -> bool:
        return [int(x) for x in re.sub(r"(\.0+)*$", "", v).split(".")]

    (a, b) = (normalize(version1), normalize(version2))
    return (a > b) - (a < b)


def import_module_by_file_name(module_file_path: str) -> Any:  # noqa: ANN401
    """Standard method of importing a Python file. Expecting absolute path."""
    module_name = os.path.basename(module_file_path)
    spec = importlib.util.spec_from_file_location(module_name, module_file_path)

    if spec is None:
        raise RuntimeError(f"Expected module file named {module_file_path}")

    ImportedModule = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(ImportedModule)

    return ImportedModule


def locate_class_in_module(Module: Any, DesiredClass: Any) -> Any:  # noqa: ANN401
    """Given a module and a class, this function will return the subclass of DesiredClass found in Module.

    It gives preference to classes that are defined in the module itself.
    This means that if you have an import that subclasses DesiredClass, it will be picked unless
    there is a class defined in the module that subclasses DesiredClass.

    In this hypothetical class hierarchy, GrandChildClass would be picked
    --------------      ------------      -----------------
    |DesiredClass|  ->  |ChildClass|  ->  |GrandChildClass|
    --------------      ------------      -----------------
    """
    DesiredClassInstance = None
    # Pull out the contents of the module that was provided
    module_contents = dir(Module)
    # Filter through the Module, we're only looking for classes.
    classList = [getattr(Module, obj) for obj in module_contents if inspect.isclass(getattr(Module, obj))]

    for _class in classList:
        # Classes that the module import show up in this list too so we need
        # to make sure it's an INSTANCE of DesiredClass, not DesiredClass itself!
        # if multiple instances are found in the same class hierarchy, pick the
        # most specific one. If multiple instances are found belonging to different
        # class hierarchies, raise an error.
        if _class is not DesiredClass and issubclass(_class, DesiredClass):
            if (DesiredClassInstance is None) or issubclass(_class, DesiredClassInstance):
                DesiredClassInstance = _class
            elif not issubclass(DesiredClassInstance, _class):
                raise RuntimeError(f"Multiple instances were found:\n\t{DesiredClassInstance}\n\t{_class}")
    return DesiredClassInstance


def RemoveTree(dir_path: str, ignore_errors: bool = False) -> None:
    """Helper for removing a directory.

    On error try to change file attributes.  Also adds retry logic.

    Args:
        dir_path: Path to directory to remove.
        ignore_errors: ignore errors during removal
    """
    # Allows for finding paths over the 260 character limit, even if the registry key (LongPathsEnabled)
    # Is not enabled.
    if sys.platform.startswith("win"):
        dir_path = "\\\\?\\" + os.path.abspath(dir_path)

    def _remove_readonly(func: Callable, path: str, _: any) -> None:
        """Private function to attempt to change permissions on file/folder being deleted."""
        os.chmod(path, stat.S_IWRITE)
        func(path)

    for _ in range(3):  # retry up to 3 times
        try:
            shutil.rmtree(dir_path, ignore_errors=ignore_errors, onerror=_remove_readonly)
        except OSError as err:
            logging.warning(f"Failed to fully remove {dir_path}: {err}")
        else:
            break
    else:
        raise RuntimeError(f"Failed to remove {dir_path}")


def PrintByteList(
    ByteList: bytearray,
    IncludeAscii: bool = True,
    IncludeOffset: bool = True,
    IncludeHexSep: bool = True,
    OffsetStart: int = 0,
    **kwargs: Any,  #
) -> None:
    """Print a byte array as hex and optionally output ascii as well as offset within the buffer."""
    outfs = kwargs.get("outfs", sys.stdout)
    kwargs["include_ascii"] = IncludeAscii
    kwargs["include_hex_sep"] = IncludeHexSep
    kwargs["include_offset"] = IncludeOffset

    warn(
        "This function is being replaced by hexdump, if you rely on this behavior switch to hexdump", DeprecationWarning
    )
    hexdump(ByteList, offset_start=OffsetStart, outfs=outfs, **kwargs)


def hexdump(byte_list: bytearray, offset_start: int = 0, outfs: BytesIO = sys.stdout, **kwargs: Any) -> None:  #
    """Print a byte array as hex and optionally output ascii as well as offset within the buffer.

    Args:
        byte_list: byte array to print
        offset_start: offset to print to the side of the hexdump
        outfs: output file stream to print to
        kwargs: keyword arguments expanded below.

    Keyword Arguments:
        include_ascii (bool): Option (Default: True) to include ascii
        include_offset (bool): Option (Default: True) to include the offset
        include_hex_sep (bool): Option (Default: True) to include the hex seperator

    Returns:
        None
    """
    include_ascii = kwargs.get("include_ascii", True)
    include_offset = kwargs.get("include_offset", True)
    include_hex_sep = kwargs.get("include_hex_sep", True)

    ascii_string = ""
    index = 0

    for index, byte in enumerate(byte_list):
        # Start of New Line
        if index % 16 == 0 and include_offset:
            outfs.write(f"{(index + offset_start):#04x} -")

        # Midpoint of a Line
        if index % 16 == 8 and include_hex_sep:
            outfs.write(" -")

        # Print As Hex Byte
        outfs.write(f" {byte:#04x}")

        # Prepare to Print As Ascii
        if byte < 0x20 or byte > 0x7E:
            ascii_string += "."
        else:
            ascii_string += f"{chr(byte)}"

        # End of Line
        if index % 16 == 15:
            if include_ascii:
                outfs.write(f" {ascii_string} ")
            ascii_string = ""
            outfs.write("\n")

    # Done - Lets check if we have partial
    if index % 16 != 15:
        # Lets print any partial line of ascii
        if include_ascii and ascii_string != "":
            # Pad out to the correct spot

            while index % 16 != 15:
                outfs.write("     ")
                if index % 16 == 7:  # account for the - symbol in the hex dump
                    if include_offset:
                        outfs.write("  ")
                index += 1
            # print the ascii partial line
            outfs.write(f" {ascii_string} ")
        # print a single newline so that next print will be on new line
        outfs.write("\n")


def export_c_type_array(buffer_fs: BytesIO, variable_name: str, out_fs: StringIO, **kwargs: Any) -> None:  #
    """Converts a given binary file to a UEFI typed C style array.

    Args:
        buffer_fs: buffer file stream to turn into a C style array
        variable_name: variable name to use for the C style array
        out_fs: output filestream to write to
        kwargs: keyword arguments expanded below.

    Keyword Arguments:
        data_type (str): The datatype of the array (Default: UINT8)
        length_data_type (str): The datatype of the length field (Default: UINTN)
        is_array (bool): if true includes '[]' (Default: True)
        bytes_per_row (int): number of bytes to include per row (Default: 16)
        indent (str): the characters to use for indention (Default: '  ' (2 spaces))
        length_variable_name (str): name to use for the length variable
        include_ascii (bool): includes a ascii comment to side of hex
        include_length (bool): includes length in the decleration of array
            (ex. "UINT8 TestVariable[13] = {")

    Return:
         None

    Raises:
        ValueError: Binary file length was 0
    """
    data_type = kwargs.get("data_type", "UINT8")
    length_data_type = kwargs.get("length_data_type", "UINTN")
    bytes_per_row = kwargs.get("bytes_per_row", 16)
    indent = kwargs.get("indent", "    ")
    include_ascii = kwargs.get("include_ascii", True)
    include_length = kwargs.get("include_length", True)
    length_variable_name = kwargs.get("length_variable_name", None)

    start = buffer_fs.tell()
    buffer_fs.seek(0, 2)
    end = buffer_fs.tell()
    buffer_fs.seek(start)
    length = end - start

    # for some reason os.linesep causes twice the amount of desired newlines
    newline = "\n"

    if length == 0:
        raise ValueError("Binary file length was 0")

    length_value = ""
    if include_length:
        length_value = str(length)

    out_fs.write(f"{data_type} {variable_name}[{length_value}] = {{")

    ascii_string = ""
    i = 0
    byte = ""

    for i, byte in enumerate(buffer_fs.read()):
        if i % bytes_per_row == 0:
            if i != 0 and include_ascii:
                # An ending space is required because the '\' character will cause a line cont.
                out_fs.write(f" // {ascii_string} ")
                ascii_string = ""
            out_fs.write(f"{newline}{indent}")

        out_fs.write(f"{byte:#04x}")

        if byte < 0x20 or byte > 0x7E:
            ascii_string += "."
        else:
            ascii_string += f"{chr(byte)}"

        if i != length - 1:
            out_fs.write(", ")

    # pad out the remaining space
    if include_ascii:
        # bytes_per_row - 1 because indexes at 0
        # subtract the number of bytes we printed
        # now we know how many bytes we could have printed
        potential_bytes = (bytes_per_row - 1) - (i % bytes_per_row)

        # pad out the number of bytes by our byte length
        # use whatever was left over in byte
        byte_length = len(f" {byte:#04x},")

        # pad out the the line
        out_fs.write(" " * potential_bytes * byte_length)

        # make up for the trailing ',' and print a comment
        # An ending space is required because the '\' character will cause a line cont.
        out_fs.write(f"   // {ascii_string} ")

    out_fs.write(f"{newline}}};{newline*2}")

    if length_variable_name:
        out_fs.write(f"{length_data_type} {length_variable_name} = sizeof {variable_name};{newline*2}")



## @file
# Module that encodes and decodes a EFI_FIRMWARE_IMAGE_AUTHENTICATION with
# certificate data and payload data.
#
# Copyright (c) 2018 - 2019, Intel Corporation. All rights reserved.<BR>
# Copyright (c) Microsoft Corporation
# SPDX-License-Identifier: BSD-2-Clause-Patent
#

"""Module for encoding and decoding EFI_FIRMWARE_IMAGE_AUTHENTICATION with certificate data and payload data."""

import struct
from typing import IO

from edk2toollib.uefi.edk2.fmp_payload_header import FmpPayloadHeaderClass
from edk2toollib.uefi.wincert import WinCertUefiGuid


class FmpAuthHeaderClass(object):
    r"""An object representing an EFI_FIRMWARE_IMAGE_AUTHENTICATION.

    Can parse or produce an EFI_FIRMWARE_IMAGE_AUTHENTICATION structure/byte buffer.

    Attributes:
        MonotonicCount (int):       It is included in the signature of AuthInfo. It is used to ensure freshness/no
                                    replay. It is incremented during each firmware image operation.
        AuthInfo (WinCertUefiGuid): Provides the authorization for the firmware image operations.
        Payload (str):              string representing payload as bytes (i.e. b'\x01\x00\x03')
        FmpPayloadHeader (FmpPayloadHeaderClass): Header for the payload

    ```
    typedef struct {
        UINT64                                  MonotonicCount;
        WIN_CERTIFICATE_UEFI_GUID               AuthInfo;
    } EFI_FIRMWARE_IMAGE_AUTHENTICATION;
    ```
    """

    _MonotonicCountFormat = "<Q"
    _MonotonicCountSize = struct.calcsize(_MonotonicCountFormat)

    def __init__(self) -> "FmpAuthHeaderClass":
        """Inits an empty object."""
        self.MonotonicCount = 0
        self.AuthInfo = WinCertUefiGuid()
        self.Payload = b""
        self.FmpPayloadHeader = None

    def Encode(self) -> bytes:
        r"""Serializes the Auth header + AuthInfo + Payload/FmpPayloadHeader.

        Returns:
            (bytes): string representing packed data as bytes (i.e. b'\x01\x00\x03')
        """
        FmpAuthHeader = struct.pack(self._MonotonicCountFormat, self.MonotonicCount)

        if self.FmpPayloadHeader is not None:
            return FmpAuthHeader + self.AuthInfo.Encode() + self.FmpPayloadHeader.Encode()
        else:
            return FmpAuthHeader + self.AuthInfo.Encode() + self.Payload

    def Decode(self, Buffer: IO) -> bytes:
        """Loads data into the Object by parsing a buffer.

        Args:
            Buffer (obj): Buffer containing the data

        Returns:
            (str): string of binary representing the payload

        Raises:
            (ValueError): Invalid Buffer
        """
        if len(Buffer) < self._MonotonicCountSize:
            raise ValueError
        (MonotonicCount,) = struct.unpack(self._MonotonicCountFormat, Buffer[: self._MonotonicCountSize])
        self.MonotonicCount = MonotonicCount

        self.Payload = self.AuthInfo.Decode(Buffer[self._MonotonicCountSize :])
        if len(self.Payload) > 0:
            self.FmpPayloadHeader = FmpPayloadHeaderClass()
            self.FmpPayloadHeader.Decode(self.Payload)
        return self.Payload

    def IsSigned(self, Buffer: IO) -> bool:
        """Parses the buffer and returns if the Cert is signed or not.

        Returns:
            (bool): True if signed
            (bool): False if invalid buffer
            (bool): False if not signed
        """
        if len(Buffer) < self._MonotonicCountSize:
            return False

        auth_info = WinCertUefiGuid(Buffer[self._MonotonicCountSize :])
        if auth_info.CertType != WinCertUefiGuid._EFI_CERT_TYPE_PKCS7_GUID.bytes_le:
            return False
        return True

    def DumpInfo(self) -> None:
        """Prints object to console."""
        print(
            "EFI_FIRMWARE_IMAGE_AUTHENTICATION.MonotonicCount                = {MonotonicCount:016X}".format(
                MonotonicCount=self.MonotonicCount
            )
        )
        self.AuthInfo.DumpInfo()
        print(
            "sizeof (Payload)                                                = {Size:08X}".format(
                Size=len(self.Payload)
            )
        )
        if self.FmpPayloadHeader is not None:
            self.FmpPayloadHeader.DumpInfo()



# @file guid_list
#
# Simple list of GuidListEntry objects parsed from edk2 specific files.
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Simple list of GuidListEntry objects parsed from edk2 specific files."""

import logging
import os
from typing import IO

from edk2toollib.gitignore_parser import parse_gitignore_lines
from edk2toollib.uefi.edk2.parsers.dec_parser import DecParser
from edk2toollib.uefi.edk2.parsers.inf_parser import InfParser


class GuidListEntry:
    """A object representing a Guid.

    Attributes:
        name (str): name of guid
        guid (str): registry format guid in string format
        filepath (str): absolute path to file where this guid was found
    """

    def __init__(self, name: str, guid: str, filepath: str) -> "GuidListEntry":
        """Create GuidListEntry for later review and compare.

        Args:
            name (str): name of guid
            guid (str): registry format guid in string format
            filepath (str): absolute path to file where this guid was found
        """
        self.name = name
        self.guid = guid
        self.absfilepath = filepath

    def __str__(self) -> str:
        """String representation of the guid."""
        return f"GUID: {self.guid} NAME: {self.name} FILE: {self.absfilepath}"


class GuidList:
    """Static class for returning Guids."""

    @staticmethod
    def guidlist_from_filesystem(folder: str, ignore_lines: list = list()) -> list:
        """Create a list of GuidListEntry from files found in the file system.

        Args:
            folder (str): path string to root folder to walk
            ignore_lines (list): list of gitignore syntax to ignore files and folders

        Returns:
            (list[GuidListEntry]): guids
        """
        guids = []
        ignore = parse_gitignore_lines(ignore_lines, os.path.join(folder, "nofile.txt"), folder)
        for root, dirs, files in os.walk(folder):
            for d in dirs[:]:
                fullpath = os.path.join(root, d)
                if ignore(fullpath):
                    logging.debug(f"Ignore folder: {fullpath}")
                    dirs.remove(d)

            for name in files:
                fullpath = os.path.join(root, name)
                if ignore(fullpath):
                    logging.debug(f"Ignore file: {fullpath}")
                    continue

                new_guids = GuidList.parse_guids_from_edk2_file(fullpath)
                guids.extend(new_guids)
        return guids

    @staticmethod
    def parse_guids_from_edk2_file(filename: str) -> list:
        """Parse edk2 files for guids.

        Args:
            filename (str): abspath to dec file

        Returns:
            (list[GuidListEntry]): guids
        """
        if filename.lower().endswith(".dec"):
            with open(filename, "r") as f:
                return GuidList.parse_guids_from_dec(f, filename)
        elif filename.lower().endswith(".inf"):
            return GuidList.parse_guids_from_inf(filename)
        else:
            return []

    @staticmethod
    def parse_guids_from_dec(stream: IO, filename: str) -> list:
        """Find all guids in a dec file contents contained with stream.

        Args:
            stream: lines of dec file content
            filename: abspath to dec file

        Returns:
            (list[GuidListEntry]): Guids
        """
        results = []
        dec = DecParser()
        dec.ParseStream(stream)
        for p in dec.Protocols:
            results.append(GuidListEntry(p.name, str(p.guid).upper(), filename))
        for p in dec.PPIs:
            results.append(GuidListEntry(p.name, str(p.guid).upper(), filename))
        for p in dec.Guids:
            results.append(GuidListEntry(p.name, str(p.guid).upper(), filename))

        try:
            results.append(GuidListEntry(dec.Dict["PACKAGE_NAME"], dec.Dict["PACKAGE_GUID"], filename))
        except Exception:
            logging.warning("Failed to find Package Guid from dec file: " + filename)
        return results

    @staticmethod
    def parse_guids_from_inf(filename: str) -> list:
        """Find the module guid in an Edk2 inf file.

        Args:
            filename (str): abspath to inf file

        Returns:
            (list[GuidListEntry]): Guids
        """
        inf = InfParser()
        inf.ParseFile(filename)
        try:
            return [GuidListEntry(inf.Dict["BASE_NAME"], inf.Dict["FILE_GUID"].upper(), filename)]
        except Exception:
            logging.warning("Failed to find info from INF file: " + filename)
        return []



##
# Handle basic logging with color via ANSI commands
# Will call into win32 commands as needed when needed
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Handle basic logging with color via ANSI commands.

Will call into win32 commands as needed when needed
"""

import logging
import re
from typing import IO, Callable, Optional

from edk2toollib.utility_functions import GetHostInfo

try:
    # try to import windows types from winDLL
    import ctypes
    from ctypes import LibraryLoader

    windll = LibraryLoader(ctypes.WinDLL)
    from ctypes import wintypes
except (AttributeError, ImportError):
    # if we run into an exception (ie on unix or linux)
    windll = None

    def SetConsoleTextAttribute() -> None:
        """Create blank lambda for when on unix / linux."""
        None

    def winapi_test() -> None:
        """Create blank lambda for when on unix / linux."""
        None

else:
    # if we don't raise an exception when we import windows types
    # then execute this but don't catch an exception if raised
    from ctypes import Structure, byref

    # inspired by https://github.com/tartley/colorama/
    class CONSOLE_SCREEN_BUFFER_INFO(Structure):
        """Object representing a console screen buffer."""

        COORD = wintypes._COORD
        """struct in wincon.h."""
        _fields_ = [
            ("dwSize", COORD),
            ("dwCursorPosition", COORD),
            ("wAttributes", wintypes.WORD),
            ("srWindow", wintypes.SMALL_RECT),
            ("dwMaximumWindowSize", COORD),
        ]

        def __str__(self) -> str:
            """String representation of the console screen buffer."""
            return "(%d,%d,%d,%d,%d,%d,%d,%d,%d,%d,%d)" % (
                self.dwSize.Y,
                self.dwSize.X,
                self.dwCursorPosition.Y,
                self.dwCursorPosition.X,
                self.wAttributes,
                self.srWindow.Top,
                self.srWindow.Left,
                self.srWindow.Bottom,
                self.srWindow.Right,
                self.dwMaximumWindowSize.Y,
                self.dwMaximumWindowSize.X,
            )

    class Win32Console(object):
        """A simple wrapper around the few methods calls to windows."""

        _GetConsoleScreenBufferInfo = windll.kernel32.GetConsoleScreenBufferInfo
        _SetConsoleTextAttribute = windll.kernel32.SetConsoleTextAttribute
        _SetConsoleTextAttribute.argtypes = [
            wintypes.HANDLE,
            wintypes.WORD,
        ]
        _SetConsoleTextAttribute.restype = wintypes.BOOL
        _GetStdHandle = windll.kernel32.GetStdHandle
        _GetStdHandle.argtypes = [
            wintypes.DWORD,
        ]
        _GetStdHandle.restype = wintypes.HANDLE

        # from winbase.h
        STDOUT = -11
        STDERR = -12

        @staticmethod
        def _winapi_test(handle: Callable) -> bool:
            csbi = CONSOLE_SCREEN_BUFFER_INFO()
            success = Win32Console._GetConsoleScreenBufferInfo(handle, byref(csbi))
            return bool(success)

        @staticmethod
        def winapi_test() -> Callable:
            """Returns the winapi_test."""
            return any(
                Win32Console._winapi_test(h)
                for h in (
                    Win32Console._GetStdHandle(Win32Console.STDOUT),
                    Win32Console._GetStdHandle(Win32Console.STDERR),
                )
            )

        @staticmethod
        def GetConsoleScreenBufferInfo(stream_id: int = STDOUT) -> CONSOLE_SCREEN_BUFFER_INFO:
            """Returns the console screen buffer info object."""
            handle = Win32Console._GetStdHandle(stream_id)
            csbi = CONSOLE_SCREEN_BUFFER_INFO()
            Win32Console._GetConsoleScreenBufferInfo(handle, byref(csbi))
            return csbi

        @staticmethod
        def SetConsoleTextAttribute(stream_id: int, attrs: int) -> bool:
            """Sets the console text attribute."""
            handle = Win32Console._GetStdHandle(stream_id)
            return Win32Console._SetConsoleTextAttribute(handle, attrs)


# from wincon.h
class WinColor(object):
    """Enum representing Windows Console colors."""

    BLACK = 0
    BLUE = 1
    GREEN = 2
    CYAN = 3
    RED = 4
    MAGENTA = 5
    YELLOW = 6
    GREY = 7
    NORMAL = 0x00  # dim text, dim background
    BRIGHT = 0x08  # bright text, dim background
    BRIGHT_BACKGROUND = 0x80  # dim text, bright background


class AnsiColor(object):
    """Defines the different codes for the ansi colors."""

    BLACK = 30
    RED = 31
    GREEN = 32
    YELLOW = 33
    BLUE = 34
    MAGENTA = 35
    CYAN = 36
    WHITE = 37
    RESET = 39
    LIGHTBLACK_EX = 90
    LIGHTRED_EX = 91
    LIGHTGREEN_EX = 92
    LIGHTYELLOW_EX = 93
    LIGHTBLUE_EX = 94
    LIGHTMAGENTA_EX = 95
    LIGHTCYAN_EX = 96
    LIGHTWHITE_EX = 97
    BG_BLACK = 40
    BG_RED = 41
    BG_GREEN = 42
    BG_YELLOW = 43
    BG_BLUE = 44
    BG_MAGENTA = 45
    BG_CYAN = 46
    BG_WHITE = 47
    BG_RESET = 49
    # These are fairly well supported, but not part of the standard.
    BG_LIGHTBLACK_EX = 100
    BG_LIGHTRED_EX = 101
    BG_LIGHTGREEN_EX = 102
    BG_LIGHTYELLOW_EX = 103
    BG_LIGHTBLUE_EX = 104
    BG_LIGHTMAGENTA_EX = 105
    BG_LIGHTCYAN_EX = 106
    BG_LIGHTWHITE_EX = 107

    @classmethod
    def __contains__(self: "AnsiColor", item: str) -> bool:
        """Verifies we contain the color."""
        if isinstance(item, str) and hasattr(self, item):
            return True
        # check if we contain the color number
        for attr_name in dir(self):
            if getattr(self, attr_name) is item:
                return True
        return False


class ColoredFormatter(logging.Formatter):
    """The formatter that outputs ANSI codes as needed."""

    AZURE_COLORS = {"CRITICAL": "section", "ERROR": "error"}

    COLORS = {
        "WARNING": AnsiColor.YELLOW,
        "INFO": AnsiColor.CYAN,
        "DEBUG": AnsiColor.BLUE,
        "CRITICAL": AnsiColor.LIGHTWHITE_EX,
        "ERROR": AnsiColor.RED,
        "STATUS": AnsiColor.GREEN,
        "PROGRESS": AnsiColor.GREEN,
        "SECTION": AnsiColor.CYAN,
    }

    def __init__(self, msg: str = "", use_azure: bool = False) -> "ColoredFormatter":
        """Inits the formatter."""
        logging.Formatter.__init__(self, msg)
        self.use_azure = use_azure

    def format(self, record: logging.LogRecord) -> str:
        """Formats the given record and returns it."""
        levelname = record.levelname
        org_message = record.msg

        if not self.use_azure and levelname in ColoredFormatter.COLORS:
            # just color the level name
            if record.levelno < logging.WARNING:
                levelname_color = get_ansi_string(ColoredFormatter.COLORS[levelname]) + levelname + get_ansi_string()
            # otherwise color the wholes message
            else:
                levelname_color = get_ansi_string(ColoredFormatter.COLORS[levelname]) + levelname
                record.msg = str(org_message) + get_ansi_string()
            record.levelname = levelname_color

        if self.use_azure and levelname in ColoredFormatter.AZURE_COLORS:
            levelname_color = "##[" + ColoredFormatter.AZURE_COLORS[levelname] + "]"
            record.levelname = levelname_color

        result = logging.Formatter.format(self, record)

        record.levelname = levelname
        record.msg = org_message
        return result


def get_ansi_string(color: AnsiColor = AnsiColor.RESET) -> str:
    """Returns the string formatted ANSI command for the specific color."""
    CSI = "\033["
    colors = AnsiColor()
    if color not in colors:
        color = AnsiColor.RESET
    return CSI + str(color) + "m"


class ColoredStreamHandler(logging.StreamHandler):
    """Class for logging in Color.."""

    # Control Sequence Introducer
    ANSI_CSI_RE = re.compile("\001?\033\\[((?:\\d|;)*)([a-zA-Z])\002?")

    def __init__(
        self, stream: Optional[IO] = None, strip: Optional[str] = None, convert: Optional[str] = None
    ) -> "ColoredStreamHandler":
        """Inits a Colored Stream Handler."""
        logging.StreamHandler.__init__(self, stream)
        self.on_windows = GetHostInfo().os == "Windows"
        # We test if the WinAPI works, because even if we are on Windows
        # we may be using a terminal that doesn't support the WinAPI
        # (e.g. Cygwin Terminal). In this case it's up to the terminal
        # to support the ANSI codes.
        self.conversion_supported = self.on_windows and Win32Console.winapi_test()
        self.strip = False
        # should we strip ANSI sequences from our output?
        if strip is None:
            strip = self.conversion_supported or (not self.stream.closed and not self.stream.isatty())
        self.strip = strip

        # should we should convert ANSI sequences into win32 calls?
        if convert is None:
            convert = self.conversion_supported and not self.stream.closed and self.stream.isatty()
        self.convert = convert
        self.win32_calls = None

        if stream is not None:
            self.stream = stream

        if self.on_windows:
            self.win32_calls = self.get_win32_calls()
            self._light = 0
            self._default = Win32Console.GetConsoleScreenBufferInfo(Win32Console.STDOUT).wAttributes
            self.set_attrs(self._default)
            self._default_fore = self._fore
            self._default_back = self._back
            self._default_style = self._style

    def handle(self, record: logging.LogRecord) -> bool:
        """Conditionally emit the specified logging record.

        Emission depends on filters which may have been added to the handler.
        Wrap the actual emission of the record with acquisition/release of
        the I/O thread lock. Returns whether the filter passed the record for
        emission.
        """
        rv = self.filter(record)
        if rv and record.levelno >= self.level:
            self.acquire()
            try:
                self.emit(record)
            finally:
                self.release()
        return rv

    def get_win32_calls(self) -> dict:
        """Returns a dict for converting ANSI Colors to Windows Colors."""
        if self.convert:
            return {
                AnsiColor.BLACK: (self.set_foreground, WinColor.BLACK),
                AnsiColor.RED: (self.set_foreground, WinColor.RED),
                AnsiColor.GREEN: (self.set_foreground, WinColor.GREEN),
                AnsiColor.YELLOW: (self.set_foreground, WinColor.YELLOW),
                AnsiColor.BLUE: (self.set_foreground, WinColor.BLUE),
                AnsiColor.MAGENTA: (self.set_foreground, WinColor.MAGENTA),
                AnsiColor.CYAN: (self.set_foreground, WinColor.CYAN),
                AnsiColor.WHITE: (self.set_foreground, WinColor.GREY),
                AnsiColor.RESET: (self.set_foreground, None),
                AnsiColor.LIGHTBLACK_EX: (self.set_foreground, WinColor.BLACK, True),
                AnsiColor.LIGHTRED_EX: (self.set_foreground, WinColor.RED, True),
                AnsiColor.LIGHTGREEN_EX: (self.set_foreground, WinColor.GREEN, True),
                AnsiColor.LIGHTYELLOW_EX: (self.set_foreground, WinColor.YELLOW, True),
                AnsiColor.LIGHTBLUE_EX: (self.set_foreground, WinColor.BLUE, True),
                AnsiColor.LIGHTMAGENTA_EX: (self.set_foreground, WinColor.MAGENTA, True),
                AnsiColor.LIGHTCYAN_EX: (self.set_foreground, WinColor.CYAN, True),
                AnsiColor.LIGHTWHITE_EX: (self.set_foreground, WinColor.GREY, True),
                AnsiColor.BG_BLACK: (self.set_background, WinColor.BLACK),
                AnsiColor.BG_RED: (self.set_background, WinColor.RED),
                AnsiColor.BG_GREEN: (self.set_background, WinColor.GREEN),
                AnsiColor.BG_YELLOW: (self.set_background, WinColor.YELLOW),
                AnsiColor.BG_BLUE: (self.set_background, WinColor.BLUE),
                AnsiColor.BG_MAGENTA: (self.set_background, WinColor.MAGENTA),
                AnsiColor.BG_CYAN: (self.set_background, WinColor.CYAN),
                AnsiColor.BG_WHITE: (self.set_background, WinColor.GREY),
                AnsiColor.BG_RESET: (self.set_background, None),
                AnsiColor.BG_LIGHTBLACK_EX: (self.set_background, WinColor.BLACK, True),
                AnsiColor.BG_LIGHTRED_EX: (self.set_background, WinColor.RED, True),
                AnsiColor.BG_LIGHTGREEN_EX: (self.set_background, WinColor.GREEN, True),
                AnsiColor.BG_LIGHTYELLOW_EX: (self.set_background, WinColor.YELLOW, True),
                AnsiColor.BG_LIGHTBLUE_EX: (self.set_background, WinColor.BLUE, True),
                AnsiColor.BG_LIGHTMAGENTA_EX: (self.set_background, WinColor.MAGENTA, True),
                AnsiColor.BG_LIGHTCYAN_EX: (self.set_background, WinColor.CYAN, True),
                AnsiColor.BG_LIGHTWHITE_EX: (self.set_background, WinColor.GREY, True),
            }
        return dict()

    def set_foreground(self, fore: Optional[int] = None, light: bool = False, on_stderr: bool = False) -> None:
        """Does the win32 call to set the foreground."""
        if fore is None:
            fore = self._default_fore
        self._fore = fore
        # Emulate LIGHT_EX with BRIGHT Style
        if light:
            self._light |= WinColor.BRIGHT
        else:
            self._light &= ~WinColor.BRIGHT
        self.set_console(on_stderr=on_stderr)

    def set_background(self, back: Optional[int] = None, light: bool = False, on_stderr: bool = False) -> None:
        """Does the win32 call to see the background."""
        if back is None:
            back = self._default_back
        self._back = back
        # Emulate LIGHT_EX with BRIGHT_BACKGROUND Style
        if light:
            self._light |= WinColor.BRIGHT_BACKGROUND
        else:
            self._light &= ~WinColor.BRIGHT_BACKGROUND
        self.set_console(on_stderr=on_stderr)

    def set_console(self, attrs: Optional[int] = None, on_stderr: bool = False) -> None:
        """Does the win32 call to set the console text attribute."""
        if attrs is None:
            attrs = self.get_attrs()
        handle = Win32Console.STDOUT
        if on_stderr:
            handle = Win32Console.STDERR
        Win32Console.SetConsoleTextAttribute(handle, attrs)

    def get_attrs(self) -> int:
        """Gets the current settings for the style and colors selected."""
        return self._fore + self._back * 16 + (self._style | self._light)

    def set_attrs(self, value: int) -> None:
        """Sets the attributes for the style and colors selected."""
        self._fore = value & 7
        self._back = (value >> 4) & 7
        self._style = value & (WinColor.BRIGHT | WinColor.BRIGHT_BACKGROUND)

    def write(self, text: str) -> None:
        """Writes to stream, stripping ANSI if specified."""
        if self.strip or self.convert:
            self.write_and_convert(text)
        else:
            self.write_plain_text(text)

    def write_and_convert(self, text: str) -> None:
        """Write the given text to the strip stripping and converting ANSI."""
        cursor = 0
        for match in self.ANSI_CSI_RE.finditer(text):
            start, end = match.span()
            if cursor < start:
                self.write_plain_text(text, cursor, start)
            self.convert_ansi(*match.groups())
            cursor = end

        self.write_plain_text(text, cursor, len(text))

    def write_plain_text(self, text: str, start: Optional[int] = None, end: Optional[int] = None) -> None:
        """Writes plain text to our stream."""
        if start is None:
            self.stream.write(text)
        elif start < end:
            self.stream.write(text[start:end])
        self.flush()

    def convert_ansi(self, paramstring: str, command: str) -> None:
        """Converts an ANSI command to a win32 command."""
        if self.convert:
            params = self.extract_params(command, paramstring)
            self.call_win32(command, params)

    def extract_params(self, command: str, paramstring: str) -> tuple:
        """Extracts the parameters in the ANSI command."""
        params = tuple(int(p) for p in paramstring.split(";") if len(p) != 0)
        if len(params) == 0:
            params = (0,)

        return params

    def call_win32(self, command: str, params: list) -> None:
        """Calls the win32 apis set_foreground and set_background."""
        if command == "m":
            for param in params:
                if param in self.win32_calls:
                    func_args = self.win32_calls[param]
                    func = func_args[0]
                    args = func_args[1:]
                    kwargs = dict()
                    func(*args, **kwargs)

    def emit(self, record: logging.LogRecord) -> None:
        """Logging.handler method we are overriding to emit a record."""
        try:
            if record is None:
                return
            msg = self.format(record)
            if msg is None:
                return
            self.write(str(msg))
            self.write(self.terminator)
            self.flush()
        except Exception:
            self.handleError(record)



# @file environment_table.py
# A module to run a table generator that creates or appends to a table with environment information."
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""A module to run a table generator that creates or appends to a table with environment information."""

import datetime
from typing import Any

import git

from edk2toollib.database import Environment, Session, Value
from edk2toollib.database.tables import TableGenerator
from edk2toollib.uefi.edk2.path_utilities import Edk2Path


class EnvironmentTable(TableGenerator):
    """A Workspace parser that records import environment information for a given parsing execution."""  # noqa: E501

    def __init__(self, *args: Any, **kwargs: Any) -> "EnvironmentTable":
        """Initialize the query with the specific settings."""

    def parse(self, session: Session, pathobj: Edk2Path, id: str, env: dict) -> None:
        """Parses the environment and adds the data to the table."""
        dtime = datetime.datetime.now()

        try:
            version = git.Repo(pathobj.WorkspacePath).head.commit.hexsha
        except git.InvalidGitRepositoryError:
            version = "UNKNOWN"

        entry = Environment(
            id=id,
            date=dtime,
            version=version,
            values=[Value(env_id=env, key=key, value=value) for key, value in env.items()],
        )

        session.add(entry)



## @file override_parser.py
# Contains classes to help with the parsing of INF files that
# may contain OVERRIDE information.
#
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Contains classes to help with parsing INF files that may contain OVERRIDE information."""

import datetime
import os
from typing import Optional

FORMAT_VERSION_1 = (1, 4)  # Version 1: #OVERRIDE : VERSION | PATH_TO_MODULE | HASH | YYYY-MM-DDThh-mm-ss


class OpParseError(Exception):
    """Class representing OpParseError types."""

    PE_VER = "VERSION"
    PE_PATH = "PATH"
    PE_HASH = "HASH"
    PE_DATE = "DATE"

    def __init__(self, my_type: str) -> "OpParseError":
        """Verifies type is a valid OpParseError type."""
        if my_type not in (OpParseError.PE_VER, OpParseError.PE_PATH, OpParseError.PE_HASH, OpParseError.PE_DATE):
            raise ValueError("Unknown type '%s'" % my_type)
        self.type = my_type

    def __str__(self) -> str:
        """String representation of the OpParseError type."""
        return repr(self.type)


class OverrideParser(object):
    """OverrideParser is a simple file parser for .inf files.

    OverrideParser must contain OVERRIDE data (i.e. overriding other .infs).
    Creating the object can be done by passing either a valid file path
    or a string containing the contents of an .inf file.

    Will raise an exception if the file doesn't exist or if the contents
    do not contain any OVERRIDE data.

    NOTE: There is an argument to be made that this class should actually be
          a subclass of InfParser, however, the InfParser is looking for far
          more details and has a much higher overhead. During a parser refactor,
          this should be considered.

    NOTE: There is a pattern used here where the object parses during
          instantiation. This pattern does not necessarily match the other
          parsers. The pros and cons of this should also be weighed during
          any parser refactor.
    """

    def __init__(self, file_path: Optional[str] = None, inf_contents: Optional[str] = None) -> "OverrideParser":
        """Inits and parses either a file or already parsed contents.

        Args:
            file_path: Path to an INF file
            inf_contents: Parsed lines as a string

        NOTE: Either file_path or inf_contents must be provided.
        """
        super(OverrideParser, self).__init__()

        # Make sure that at least some data is provided.
        if file_path is None and inf_contents is None:
            raise ValueError("file_path or inf_contents is required.")
        # Make sure not too much data is provided.
        if file_path is not None and inf_contents is not None:
            raise ValueError("Only provide file_path or inf_contents. (%s, %s)" % (file_path, inf_contents))
        # If a file path was provided, make sure it exists.
        if file_path is not None:
            if not os.path.isfile(file_path):
                raise ValueError("File path '%s' does not exist." % file_path)

        self.file_path = os.path.abspath(file_path) if file_path is not None else "String Buffer"

        # Set up the contents for parsing.
        parse_contents = inf_contents
        if file_path is not None:
            with open(file_path, "r") as file:
                parse_contents = file.read()
        if not parse_contents:
            raise ValueError("Failed to read contents of file '%s'." % self.file_path)

        self.override_lines = self.get_override_lines(parse_contents)

        # If no override lines were found, we're basically done here.
        if not self.override_lines:
            raise ValueError("File '%s' did not contain any override lines." % self.file_path)

        self.overrides = []
        for override_line in self.override_lines:
            try:
                self.overrides.append(self.parse_override_line(override_line["line"]))
            except OpParseError as pe:
                raise ValueError(
                    "Parse error '%s' occurred while processing line %d of '%s'."
                    % (pe, override_line["lineno"], override_line["line"])
                )

    @staticmethod
    def get_override_lines(parse_contents: str) -> list:
        """Parses contents and returns only lines that start with #OVERRIDE.

        Returns:
            (list[str]): lines starting with #OVERRIDE
        """
        parse_lines = parse_contents.split("\n")
        result = []

        for i in range(len(parse_lines)):
            if parse_lines[i].strip().upper().startswith("#OVERRIDE"):
                result.append({"lineno": i + 1, "line": parse_lines[i].strip()})

        return result

    @staticmethod
    def parse_override_line(line_contents: str) -> dict:
        """Parses an override_line.

        Args:
            line_contents (str): a line that starts with #OVERRIDE

        Returns:
            (dict): Key/Value pairs including version, original_path, current_hash, datetime

        Raises:
            (OpParseError): Failed to parse one of the keys
        """
        result = {}

        # Split the override string into pieces.
        # First the #OVERRIDE, which is separated by a :.
        # Then everything else by |.
        line_parts = line_contents.split(":")
        line_parts = [part.strip() for part in line_parts[1].split("|")]

        # Step 1: Check version and number of blocks in this entry
        try:
            result["version"] = int(line_parts[0])
        except ValueError:
            raise OpParseError(OpParseError.PE_VER)

        # Verify this is a known version and has valid number of entries
        if not ((result["version"] == FORMAT_VERSION_1[0]) and (len(line_parts) == FORMAT_VERSION_1[1])):
            raise OpParseError(OpParseError.PE_VER)

        # Step 2: Process the path to overridden module
        # Normalize the path to support different slashes.
        result["original_path"] = os.path.normpath(line_parts[1])

        # Step 3: Grep hash entry
        result["current_hash"] = line_parts[2]

        # Step 4: Parse the time of hash generation
        try:
            result["datetime"] = datetime.datetime.strptime(line_parts[3], "%Y-%m-%dT%H-%M-%S")
        except ValueError:
            raise OpParseError(OpParseError.PE_DATE)

        return result



# @file wincert.py
# Code to work with UEFI WinCert data
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##

"""Module for working with UEFI WinCert data."""

import io
import struct
import sys
import uuid
from io import BytesIO, StringIO
from typing import Optional, Union
from warnings import warn

from pyasn1.codec.der.decoder import decode as der_decode
from pyasn1_modules import rfc2315

from edk2toollib.utility_functions import hexdump


class WinCertPkcs1(object):
    """Object representing a WinCertPkcs1 struct.

    Certificate which encapsulates the RSASSA_PKCS1-v1_5 digital signature.

    Attributes:
        Hdr_dwLength (int):         The length of the entire certificate, including the length of the header, in bytes.
        Hdr_wRevision (int):        The certificate type. See WIN_CERT_TYPE_xxx for the UEFI certificate types. The UEFI
                                    specification reserves the range of certificate type values from 0x0EF0 to 0x0EFF.
        Hdr_wCertificateType (int): The following is the actual certificate. The format of the certificate depends on
                                    wCertificateType.
        hash_algorithm (uuid.UUID):  The Guid representing the hash algorithm for the Cert.
        cert_data (memoryview):      The actual Cert.

    The WIN_CERTIFICATE_UEFI_PKCS1_15 structure is derived from
    WIN_CERTIFICATE and encapsulate the information needed to
    implement the RSASSA-PKCS1-v1_5 digital signature algorithm as
    specified in RFC2437.
    typedef struct {
        WIN_CERTIFICATE Hdr;
        EFI_GUID        HashAlgorithm;
        UINT8 Signature[];
    } WIN_CERTIFICATE_EFI_PKCS1_15;

    The WIN_CERTIFICATE structure is part of the PE/COFF specification.
    typedef struct {
        UINT32  dwLength;
        UINT16  wRevision;
        UINT16  wCertificateType;
        UINT8 bCertificate[ANYSIZE_ARRAY];
    } WIN_CERTIFICATE;
    """

    STATIC_STRUCT_SIZE = 4 + 2 + 2 + 16
    EFI_HASH_SHA256 = uuid.UUID("{51AA59DE-FDF2-4EA3-BC63-875FB7842EE9}")  # EFI_HASH_SHA256 guid defined by UEFI spec

    def __init__(self, filestream: Optional[BytesIO] = None) -> "WinCertPkcs1":
        """Inits the object."""
        if filestream is None:
            self.Hdr_dwLength = WinCertPkcs1.STATIC_STRUCT_SIZE
            self.Hdr_wRevision = WinCert.REVISION
            self.Hdr_wCertificateType = WinCert.WIN_CERT_TYPE_EFI_PKCS115
            self.hash_algorithm = None
            self.cert_data = None
        else:
            self.decode(filestream)

    @property
    def HashAlgorithm(self) -> uuid.UUID:
        """Returns the HashAlgorithm."""
        warn("HashAlgorithm is deprecated. Use hash_algorithm instead.", DeprecationWarning, 2)
        return self.hash_algorithm

    @HashAlgorithm.setter
    def HashAlgorithm(self, value: uuid.UUID) -> None:
        """Sets the HashAlgorithm."""
        warn("HashAlgorithm is deprecated. Use hash_algorithm instead.", DeprecationWarning, 2)
        self.hash_algorithm = value

    @property
    def CertData(self) -> bytes:
        """Returns the CertData."""
        warn("CertData is deprecated. Use cert_data instead.", DeprecationWarning, 2)
        return self.cert_data

    @CertData.setter
    def CertData(self, value: bytes) -> None:
        """Sets the CertData."""
        warn("CertData is deprecated. Use cert_data instead.", DeprecationWarning, 2)
        self.cert_data = value

    def add_cert_data(self, fs: BytesIO) -> None:
        """Adds the Cert Data to the struct.

        Args:
            fs (io.BytesIO): filestream representing bytes

        Raises:
            (ValueError): Invalid Cert Data
            (ValueError): Missing Hash Algorithm
        """
        if self.cert_data is not None:
            raise ValueError("Certificate data already set")

        if self.hash_algorithm is None:
            raise ValueError("You must set the Hash Algorithm first")

        self.cert_data = memoryview(fs.read())
        self.Hdr_dwLength = self.Hdr_dwLength + len(self.cert_data)

    def AddCertData(self, fs: BytesIO) -> None:
        """Adds the Cert Data to the struct.

        Raises:
            (ValueError): Invalid Cert Data
            (ValueError): Missing Hash Algorithm
        """
        warn("AddCertData is deprecated, use add_cert_data instead", DeprecationWarning, 2)
        self.add_cert_data(fs)

    def set_hash_algorithm(self, hash_algorithm: uuid.UUID) -> None:
        """Sets the hash algorithm for the wincert.

        Args:
            hash_algorithm (uuid.UUID): The Guid representing the hash algorithm for the Cert.
        """
        self.hash_algorithm = hash_algorithm

    def decode(self, fs: BytesIO) -> None:
        """Populates the struct from a filestream.

        Args:
            fs: An open file

        Raises:
            (ValueError): Invalid stream
            (ValueError): Invalid stream size
        """
        if fs is None:
            raise ValueError("Invalid File stream")

        # only populate from file stream those parts that are complete in the file stream
        offset = fs.tell()
        fs.seek(0, 2)
        end = fs.tell()
        fs.seek(offset)

        if (end - offset) < WinCertPkcs1.STATIC_STRUCT_SIZE:  # size of the static header data
            raise ValueError("Invalid file stream size")

        self.Hdr_dwLength = struct.unpack("=I", fs.read(4))[0]
        self.Hdr_wRevision = struct.unpack("=H", fs.read(2))[0]
        self.Hdr_wCertificateType = struct.unpack("=H", fs.read(2))[0]
        self.hash_algorithm = uuid.UUID(bytes_le=fs.read(16))
        self.cert_data = None

        if (end - fs.tell()) < 1:
            raise ValueError("Invalid File stream. No data for signature cert data")

        if (end - fs.tell()) < (self.Hdr_dwLength - WinCertPkcs1.STATIC_STRUCT_SIZE):
            raise ValueError("Invalid file stream size")

        self.cert_data = memoryview(fs.read(self.Hdr_dwLength - WinCertPkcs1.STATIC_STRUCT_SIZE))

    def PopulateFromFileStream(self, fs: BytesIO) -> None:
        """Populates the struct from a filestream.

        Args:
            fs: an open file

        Raises:
            (ValueError): Invalid stream
            (ValueError): Invalid stream size
        """
        warn("PopulateFromFileStream is deprecated, use decode instead", DeprecationWarning, 2)
        return self.decode(fs)

    def print(self, out_fs: Optional[BytesIO] = sys.stdout) -> None:
        """Prints the struct to the console."""
        out_fs.write("\n-------------------- WinCertPKCS115 ---------------------\n")
        out_fs.write(f"  Hdr_dwLength:         0x{self.Hdr_dwLength:0X}\n")
        out_fs.write(f"  Hdr_wRevision:        0x{self.Hdr_wRevision:0X}\n")
        out_fs.write(f"  Hdr_wCertificateType: 0x{self.Hdr_wCertificateType:0X}\n")
        out_fs.write(f"  Hash Guid:            {str(self.hash_algorithm)}\n")
        out_fs.write("  cert_data:             \n")
        cdl = self.cert_data.tolist()
        hexdump(cdl, out_fs=out_fs)

    def Print(self, out_fs: Optional[BytesIO] = sys.stdout) -> None:
        """Prints the struct to the console."""
        warn("Print is deprecated, use print instead", DeprecationWarning, 2)
        self.print(out_fs)

    def encode(self) -> bytes:
        r"""Serializes the object.

        Returns:
            (bytes): string representing packed data as bytes (i.e. b'\x01\x00\x03')
        """
        output = b""
        output += struct.pack("=I", self.Hdr_dwLength)
        output += struct.pack("=H", self.Hdr_wRevision)
        output += struct.pack("=H", self.Hdr_wCertificateType)
        output += self.hash_algorithm.bytes_le
        output += self.cert_data

        return output

    def write(self, fs: BytesIO) -> None:
        r"""Writes an serialized object to a filestream.

        Args:
            fs (obj): an open file

        """
        fs.write(self.encode())

    def Write(self, fs: BytesIO) -> None:
        r"""Serializes the object.

        Args:
            fs: an open file
        """
        warn("Write is deprecated, use encode instead", DeprecationWarning, 2)
        self.write(fs)


class WinCertUefiGuid(object):
    """Object representing a Certificate which encapsulates a GUID-specific digital signature.

    Attributes:
        Hdr_dwLength (int):         The length of the entire certificate, including the length of the header, in bytes.
        Hdr_wRevision (int):        The certificate type. See WIN_CERT_TYPE_xxx for the UEFI certificate types. The UEFI
                                    specification reserves the range of certificate type values from 0x0EF0 to 0x0EFF.
        Hdr_wCertificateType (int): The following is the actual certificate. The format of the certificate depends on
                                    wCertificateType.
        cert_data (memoryview):      The actual Cert.

    typedef struct {
        WIN_CERTIFICATE   Hdr;
        EFI_GUID          cert_type;
        UINT8            cert_data[1];
    } WIN_CERTIFICATE_UEFI_GUID;

    The WIN_CERTIFICATE structure is part of the PE/COFF specification.
    typedef struct {
        UINT32  dwLength;
        UINT16  wRevision;
        UINT16  wCertificateType;
        UINT8 bCertificate[ANYSIZE_ARRAY];
    } WIN_CERTIFICATE;
    """

    _StructFormat = "<L2H16s"
    _StructSize = struct.calcsize(_StructFormat)

    _EFI_CERT_TYPE_PKCS7_GUID = uuid.UUID("4aafd29d-68df-49ee-8aa9-347d375665a7")

    # Preserved for back compat.
    STATIC_STRUCT_SIZE = _StructSize
    PKCS7Guid = _EFI_CERT_TYPE_PKCS7_GUID

    def __init__(self, in_data: Optional[BytesIO] = None) -> "WinCertUefiGuid":
        """Inits the object."""
        self.Hdr_dwLength = self._StructSize
        self.Hdr_wRevision = WinCert.REVISION
        self.Hdr_wCertificateType = WinCert.WIN_CERT_TYPE_EFI_GUID
        self.cert_type = self._EFI_CERT_TYPE_PKCS7_GUID
        self.cert_data = b""

        if in_data is not None:
            self.decode(in_data)

    @property
    def CertData(self) -> uuid.UUID:
        """Returns the CertData."""
        warn("CertData is deprecated. Use cert_data instead.", DeprecationWarning, 2)
        return self.cert_data

    @CertData.setter
    def CertData(self, value: uuid.UUID) -> None:
        """Sets the CertData."""
        warn("CertData is deprecated. Use cert_data instead.", DeprecationWarning, 2)
        self.cert_data = value

    @property
    def CertType(self) -> uuid.UUID:
        """Returns the CertType."""
        warn("CertType is deprecated. Use cert_type instead.", DeprecationWarning, 2)
        return self.cert_type

    @CertType.setter
    def CertType(self, value: uuid.UUID) -> None:
        """Sets the CertType."""
        warn("CertType is deprecated. Use cert_type instead.", DeprecationWarning, 2)
        self.cert_type = value

    def get_length(self) -> int:
        """Returns the length of the WinCertUefiGuid and it's data."""
        return self.Hdr_dwLength

    def Encode(self) -> str:
        r"""Serializes the object.

        Returns:
            (str): string representing packed data as bytes (i.e. b'\x01\x00\x03')

        Raises:
            (ValueError): Invalid Revision
            (ValueError): Invalid Cert Type
        """
        warn("Encode is deprecated, use encode instead", DeprecationWarning, 2)
        return self.encode()

    def encode(self) -> str:
        r"""Serializes the object.

        Returns:
            (str): string representing packed data as bytes (i.e. b'\x01\x00\x03')

        Raises:
            (ValueError): Invalid Revision
            (ValueError): Invalid Cert Type
        """
        if self.Hdr_wRevision != WinCert.REVISION:
            raise ValueError("Invalid revision")
        if self.Hdr_wCertificateType != WinCert.WIN_CERT_TYPE_EFI_GUID:
            raise ValueError("Invalid win_certificate type (WIN_CERT_TYPE_EFI_GUID)")
        if self.cert_type != self._EFI_CERT_TYPE_PKCS7_GUID:
            raise ValueError("Invalid certificate type (EFI_CERT_TYPE_PKCS7_GUID)")
        self.Hdr_dwLength = self._StructSize + len(self.cert_data)

        win_cert_header = struct.pack(
            self._StructFormat,
            self.Hdr_dwLength,
            self.Hdr_wRevision,
            self.Hdr_wCertificateType,
            self.cert_type.bytes_le,
        )

        return win_cert_header + self.cert_data

    def Decode(self, Buffer: BytesIO) -> bytes:
        """Loads the struct with values from a buffer.

        Args:
            Buffer: Buffer containing serialized data

        Returns:
            (bytes): Any remaining buffer

        Raises:
            (ValueError): Invalid Buffer
            (ValueError): Invalid dwlength
            (ValueError): Invalid Revision
            (ValueError): Invalid Cert Type
        """
        warn("Decode is deprecated, use decode instead", DeprecationWarning, 2)
        return self.decode(Buffer)

    def decode(self, in_data: BytesIO) -> bytes:
        """Loads the struct with values from a buffer.

        Args:
            in_data: in_data containing serialized data

        Returns:
            (bytes): Any remaining buffer

        Raises:
            (ValueError): Invalid Buffer
            (ValueError): Invalid dwLength
            (ValueError): Invalid Revision
            (ValueError): Invalid Cert Type
            (ValueError): Invalid datatype provided

        """
        if hasattr(in_data, "seek"):  # Filestream like object
            return self._from_filestream(in_data)
        elif hasattr(in_data, "decode"):  # Bytes like object
            return self._from_buffer(in_data)
        else:
            raise ValueError(
                f"Invalid datatype provided: {type(in_data)}, data may only be of type filestream or bytes"
            )

    def _from_buffer(self, buffer: BytesIO) -> bytes:
        """Loads the struct with values from a buffer.

        Args:
            buffer: Buffer containing serialized data

        Returns:
            (BytesIO): Any remaining buffer

        Raises:
            (ValueError): Invalid buffer
            (ValueError): Invalid dwLength
            (ValueError): Invalid Revision
            (ValueError): Invalid Cert Type
        """
        if len(buffer) < self._StructSize:
            raise ValueError
        (dwLength, wRevision, wCertificateType, cert_type) = struct.unpack(
            self._StructFormat, buffer[0 : self._StructSize]
        )
        if dwLength < self._StructSize:
            raise ValueError
        if wRevision != WinCert.REVISION:
            raise ValueError
        if wCertificateType != WinCert.WIN_CERT_TYPE_EFI_GUID:
            raise ValueError
        if cert_type != self._EFI_CERT_TYPE_PKCS7_GUID.bytes_le:
            raise ValueError
        self.Hdr_dwLength = dwLength
        self.Hdr_wRevision = wRevision
        self.Hdr_wCertificateType = wCertificateType
        self.cert_type = uuid.UUID(bytes_le=cert_type)
        self.cert_data = buffer[self._StructSize : self.Hdr_dwLength]

        # Return the remaining buffer, if any exists.
        return buffer[self.Hdr_dwLength :]

    def _from_filestream(self, fs: StringIO) -> bytes:
        """Un-serialized from a filestream.

        Args:
            fs (obj): Already opened file

        Raises:
            (ValueError): Invalid fs
            (ValueError): Invalid size
        """
        if fs is None:
            raise ValueError

        # Determine the end of the stream.
        current = fs.tell()
        end = fs.seek(0, io.SEEK_END)
        fs.seek(current)

        # Make sure that we can at least parse the size field.
        field_string = "<I"
        field_size = struct.calcsize(field_string)
        if (end - current) < field_size:
            raise ValueError

        # Parse the size field.
        (buffer_size,) = struct.unpack(field_string, fs.read(field_size))

        if (end - current) < buffer_size:
            raise ValueError

        fs.seek(current)
        object_buffer = fs.read(buffer_size)

        return self._from_buffer(object_buffer)

    def AddCertData(self, in_data: BytesIO) -> None:
        """Adds the Cert Data to the struct.

        Args:
            in_data (obj): Data to read.
        """
        warn("AddCertData is deprecated, use add_cert_data instead", DeprecationWarning, 2)
        return self.add_cert_data(in_data)

    def add_cert_data(self, in_data: BytesIO) -> None:
        """Adds the Cert Data to the struct.

        Args:
            in_data (obj): Data to read.
        """
        # if the data is already set, let's subtract it from the length before adding the new data
        if self.cert_data is not None:
            self.Hdr_dwLength -= len(self.cert_data)

        # Account for back compat. Behave differently for file streams.
        if hasattr(in_data, "seek"):
            self.cert_data = in_data.read()
        elif hasattr(in_data, "decode"):
            self.cert_data = in_data
        else:
            raise ValueError(
                f"Invalid datatype provided: {type(in_data)}, data may only be of type filestream or bytes"
            )

        self.Hdr_dwLength = self.Hdr_dwLength + len(self.cert_data)

    def PopulateFromFileStream(self, fs: BytesIO) -> bytes:
        """Un-serialized from a filestream.

        Args:
            fs: Already opened file

        Raises:
            (ValueError): Invalid fs
            (ValueError): Invalid size
        """
        warn("PopulateFromFileStream is deprecated, use decode instead", DeprecationWarning, 2)
        return self.decode(fs)

    def get_certificate(self) -> bytes:
        """Returns certificate data, if certificate data exists."""
        return self.cert_data

    def GetCertificate(self) -> bytes:
        """Returns certificate data, if certificate data exists."""
        warn("GetCertificate is deprecated, use get_certificate instead", DeprecationWarning, 2)
        return self.get_certificate()

    def print(self, outfs: StringIO = sys.stdout) -> None:
        """Prints struct to console."""
        self.dump_info(outfs)

    def Print(self, outfs: StringIO = sys.stdout) -> None:
        """Prints struct to console."""
        warn("Print() is deprecated, use print() instead.", DeprecationWarning, 2)
        self.dump_info(outfs)

    def dump_info(self, outfs: StringIO = sys.stdout) -> None:
        """Prints struct to a file stream."""
        outfs.write("\n-------------------- WIN_CERTIFICATE ---------------------\n")
        outfs.write(f"WIN_CERTIFICATE.dwLength         = {self.Hdr_dwLength:08X}\n")
        outfs.write(f"WIN_CERTIFICATE.wRevision        = {self.Hdr_wRevision:04X}\n")
        outfs.write(f"WIN_CERTIFICATE.wCertificateType = {self.Hdr_wCertificateType:04X}\n")
        outfs.write(f"WIN_CERTIFICATE_UEFI_GUID.cert_type             = {str(self.cert_type).upper()}\n")
        outfs.write(f"sizeof (WIN_CERTIFICATE_UEFI_GUID.cert_data)    = {len(self.cert_data):08X}\n")

        outfs.write("\n------------------- CERTIFICATE DATA ---------------------\n")

        # Technically the signature could be wrapped in a ContentInfo
        try:
            content_info, _ = der_decode(self.cert_data, asn1Spec=rfc2315.ContentInfo())

            outfs.write(str(content_info))
        except Exception:
            # But usually its not
            try:
                signed_data, _ = der_decode(self.cert_data, asn1Spec=rfc2315.SignedData())

                outfs.write(str(signed_data))
            except Exception as exc:
                raise ValueError("Unable to decode cert_data") from exc

    def DumpInfo(self, outfs: StringIO = sys.stdout) -> None:
        """Prints struct to a file stream."""
        warn("DumpInfo is deprecated, use dump_info instead", DeprecationWarning, 2)
        self.dump_info(outfs)

    def write(self, fs: BytesIO) -> None:
        """Writes the struct to a filestream."""
        fs.write(self.encode())

    def Write(self, fs: BytesIO) -> None:
        """Writes the struct to a filestream."""
        warn("Write is deprecated, use write instead", DeprecationWarning, 2)
        fs.write(self.encode())

    def __str__(self) -> str:
        """Returns the object as a string."""
        string_repr = ""
        with io.StringIO() as f:
            self.print(f)
            f.seek(0)
            string_repr = f.read()

        return string_repr


class WinCert(object):
    """Object for generating a WinCert."""

    STATIC_STRUCT_SIZE = 8
    # WIN_CERTIFICATE.wCertificateTypes UEFI Spec defined
    WIN_CERT_TYPE_NONE = 0x0000
    WIN_CERT_TYPE_PKCS_SIGNED_DATA = 0x0002
    WIN_CERT_TYPE_EFI_PKCS115 = 0x0EF0
    WIN_CERT_TYPE_EFI_GUID = 0x0EF1
    # Revision
    REVISION = 0x200

    @staticmethod
    def factory(fs: BytesIO) -> Union[WinCertUefiGuid, WinCertPkcs1]:
        """Generates a specific Cert Type depending on parsed Hdr_wCertificationType from the fs.

        Args:
            fs: filestream

        Returns:
            (WinCertUefiGuid): if Hdr_wCertificationType == WIN_CERT_TYPE_EFI_GUID
            (WinCertPkcs1): if Hdr_wCertificationType == WIN_CERT_TYPE_EFI_PKCS115
        Raises:
            (Exception): Invalid fs
            (Exception): Invalid fs size
        """
        if fs is None:
            raise Exception("Invalid File stream")

        # only populate from file stream those parts that are complete in the file stream
        offset = fs.tell()
        fs.seek(0, 2)
        end = fs.tell()
        fs.seek(offset)

        if (end - offset) < WinCert.STATIC_STRUCT_SIZE:  # size of the static header data
            raise Exception("Invalid file stream size")
        # 1 read len
        # 2 read revision
        # 3 read cert type
        fs.seek(4, 1)  # seeking past Hdr_dwLength
        fs.seek(2, 1)  # seeking past Hdr_wRevision
        Hdr_wCertificateType = struct.unpack("=H", fs.read(2))[0]

        fs.seek(offset)

        if Hdr_wCertificateType == WinCert.WIN_CERT_TYPE_EFI_GUID:
            return WinCertUefiGuid(fs)
        elif Hdr_wCertificateType == WinCert.WIN_CERT_TYPE_EFI_PKCS115:
            return WinCertPkcs1(fs)
        else:
            return None

    #
    # this method is a factory
    #

    @staticmethod
    def Factory(fs: BytesIO) -> Union[WinCertUefiGuid, WinCertPkcs1]:
        """Generates a specific Cert Type depending on parsed Hdr_wCertificationType from the fs.

        Args:
            fs: filestream

        Returns:
            (WinCertUefiGuid): if Hdr_wCertificationType == WIN_CERT_TYPE_EFI_GUID
            (WinCertPkcs1): if Hdr_wCertificationType == WIN_CERT_TYPE_EFI_PKCS115
        Raises:
            (Exception): Invalid fs
            (Exception): Invalid fs size
        """
        warn("Factory is deprecated, use factory instead", DeprecationWarning, 2)
        return WinCert.factory(fs)



# @file dec_parser.py
# Code to help parse DEC file
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Code to help parse DEC files."""

import os
import re
from typing import IO

from edk2toollib.uefi.edk2.parsers.base_parser import HashFileParser
from edk2toollib.uefi.edk2.parsers.guid_parser import GuidParser


class LibraryClassDeclarationEntry:
    """Object representing a Library Class Declaration Entry."""

    def __init__(self, packagename: str, rawtext: str = None) -> "LibraryClassDeclarationEntry":
        """Init a library Class Declaration Entry."""
        self.path = ""
        self.name = ""
        self.package_name = packagename
        if rawtext is not None:
            self._parse(rawtext)

    def _parse(self, rawtext: str) -> None:
        """Parses the rawtext line to collect the Library Class declaration information.

        information includes (name and package root relative path).

        Args:
          rawtext (str): raw text

        NOTE: expected format of rawtext is <library class name> | <package relative path to header file>
        """
        t = rawtext.partition("|")
        self.name = t[0].strip()
        self.path = t[2].strip()


class GuidedDeclarationEntry:
    """A baseclass for declaration types that have a name and guid.

    Attributes:
        name (str): name
        guidstring (str): guid
        guid (uuid.UUID): guid
        package_name (str): packagename
    """

    PROTOCOL = 1
    PPI = 2
    GUID = 3

    def __init__(self, packagename: str, rawtext: str = None) -> "GuidedDeclarationEntry":
        """Init a protocol/Ppi/or Guid declaration entry."""
        self.name = ""
        self.guidstring = ""
        self.guid = None
        self.package_name = packagename
        if rawtext is not None:
            self._parse(rawtext)

    def _parse(self, rawtext: str) -> None:
        """Parses the name and guid of a declaration.

        Args:
          rawtext (str): raw text
        """
        t = rawtext.partition("=")
        self.name = t[0].strip()
        self.guidstring = t[2].strip()
        self.guid = GuidParser.uuid_from_guidstring(self.guidstring)
        if self.guid is None:
            raise ValueError("Could not parse guid")


class ProtocolDeclarationEntry(GuidedDeclarationEntry):
    """Object representing a Protocol Declaration Entry."""

    def __init__(self, packagename: str, rawtext: str = None) -> "ProtocolDeclarationEntry":
        """Init a protocol declaration entry."""
        super().__init__(packagename, rawtext)
        self.type = GuidedDeclarationEntry.PROTOCOL


class PpiDeclarationEntry(GuidedDeclarationEntry):
    """Object representing a Ppi Declaration Entry."""

    def __init__(self, packagename: str, rawtext: str = None) -> "GuidedDeclarationEntry":
        """Init a Ppi declaration entry."""
        super().__init__(packagename, rawtext)
        self.type = GuidedDeclarationEntry.PPI


class GuidDeclarationEntry(GuidedDeclarationEntry):
    """Object representing a Guid Declaration Entry."""

    def __init__(self, packagename: str, rawtext: str = None) -> "GuidDeclarationEntry":
        """Init a Ppi declaration entry."""
        super().__init__(packagename, rawtext)
        self.type = GuidedDeclarationEntry.GUID


class PcdDeclarationEntry:
    """Object representing a Pcd Delcaration Entry.

    Attributes:
        token_space_name (str): token space name
        name (str): name
        default_value (str): value
        type (str): type
        id (str): id
        package_name: package name
    """

    def __init__(self, packagename: str, rawtext: str = None) -> "PcdDeclarationEntry":
        """Creates a PCD Declaration Entry for one PCD."""
        self.token_space_name = ""
        self.name = ""
        self.default_value = ""
        self.type = ""
        self.id = ""
        self.package_name = packagename
        if rawtext is not None:
            self._parse(rawtext)

    def _parse(self, rawtext: str) -> None:
        """Parses the PcdDeclaration Entry for one PCD."""
        sp = rawtext.partition(".")
        self.token_space_name = sp[0].strip()

        # Regular expression pattern to match the symbol '|' that is not inside quotes
        pattern = r'\|(?=(?:(?:[^\'"]*(?:\'[^\']*\'|"[^"]*"))*[^\'"]*)$)'
        op = re.split(pattern, sp[2])

        # if it's 2 long, we need to check that it's a structured PCD
        if len(op) == 2 and op[0].count(".") > 0:
            pass
        # otherwise it needs at least 4 parts
        elif len(op) < 4:
            raise Exception(f"Too few parts: {op}")
        # but also less than 5
        elif len(op) > 5:
            raise Exception(f"Too many parts: {rawtext}")
        elif len(op) == 5 and op[4].strip() != "{":
            raise Exception(f"Too many parts: {rawtext}")

        self.name = op[0].strip()
        self.default_value = op[1].strip()
        # if we don't know what the type and id, it's because it's structured
        self.type = op[2].strip() if len(op) > 2 else "STRUCTURED_PCD"
        self.id = op[3].strip() if len(op) > 2 else "STRUCTURED_PCD"


class DecParser(HashFileParser):
    """Parses an EDK2 DEC file.

    Attributes:
        Parsed (bool): If a DEC file has been parsed or not
        Lines (list): order list of lines in the Dec file
        Dict (dict): Dict of variables set in the DEC file
        LibraryClasses (list): list of Library classes
        PPIs (list): list of PPIs
        Protocols (list): list of Protocols
        Guids (list): list of Guids
        Pcds (list): list of Pcds
        IncludePaths (list): list of IncludePaths
        PackageName (str): Package Name variable also found in Dict
        Path (str): path to the DEC file
    """

    def __init__(self) -> "DecParser":
        """Init an empty Dec Parser."""
        HashFileParser.__init__(self, "DecParser")
        self.Lines = []
        self.Parsed = False
        self.Dict = {}
        self.LibraryClasses = []
        self.PPIs = []
        self.Protocols = []
        self.Guids = []
        self.Pcds = []
        self.IncludePaths = []
        self.Path = ""
        self.PackageName = None

    def _Parse(self) -> None:
        """Parses the EDK2 DEC file."""
        InDefinesSection = False
        InLibraryClassSection = False
        InProtocolsSection = False
        InGuidsSection = False
        InPPISection = False
        InPcdSection = False
        InStructuredPcdDeclaration = False
        InIncludesSection = False

        for line in self.Lines:
            sline = self.StripComment(line)

            if sline is None or len(sline) < 1:
                continue

            if InDefinesSection:
                if sline.strip()[0] == "[":
                    InDefinesSection = False
                else:
                    if sline.count("=") == 1:
                        tokens = sline.split("=", 1)
                        self.Dict[tokens[0].strip()] = tokens[1].strip()
                        if self.PackageName is None and tokens[0].strip() == "PACKAGE_NAME":
                            self.PackageName = self.Dict["PACKAGE_NAME"]
                        continue

            elif InLibraryClassSection:
                if sline.strip()[0] == "[":
                    InLibraryClassSection = False
                else:
                    t = LibraryClassDeclarationEntry(self.PackageName, sline)
                    self.LibraryClasses.append(t)
                    continue

            elif InProtocolsSection:
                if sline.strip()[0] == "[":
                    InProtocolsSection = False
                else:
                    t = ProtocolDeclarationEntry(self.PackageName, sline)
                    self.Protocols.append(t)
                    continue

            elif InGuidsSection:
                if sline.strip()[0] == "[":
                    InGuidsSection = False
                else:
                    t = GuidDeclarationEntry(self.PackageName, sline)
                    self.Guids.append(t)
                    continue

            elif InPcdSection:
                if sline.strip()[0] == "[":
                    InPcdSection = False
                elif sline.strip()[0] == "}":
                    InStructuredPcdDeclaration = False
                else:
                    if InStructuredPcdDeclaration:
                        continue
                    t = PcdDeclarationEntry(self.PackageName, sline)
                    self.Pcds.append(t)
                    if sline.rstrip()[-1] == "{":
                        InStructuredPcdDeclaration = True
                    continue

            elif InIncludesSection:
                if sline.strip()[0] == "[":
                    InIncludesSection = False
                else:
                    self.IncludePaths.append(sline.strip())
                    continue

            elif InPPISection:
                if sline.strip()[0] == "[":
                    InPPISection = False
                else:
                    t = PpiDeclarationEntry(self.PackageName, sline)
                    self.PPIs.append(t)
                    continue

            # check for different sections
            if sline.strip().lower().startswith("[defines"):
                InDefinesSection = True

            elif sline.strip().lower().startswith("[libraryclasses"):
                InLibraryClassSection = True

            elif sline.strip().lower().startswith("[protocols"):
                InProtocolsSection = True

            elif sline.strip().lower().startswith("[guids"):
                InGuidsSection = True

            elif sline.strip().lower().startswith("[ppis"):
                InPPISection = True

            elif sline.strip().lower().startswith("[pcd"):
                InPcdSection = True

            elif sline.strip().lower().startswith("[includes"):
                InIncludesSection = True

        self.Parsed = True

    def ParseStream(self, stream: IO) -> None:
        """Parse the supplied IO as a DEC file.

        Args:
            stream (IOBase): a file-like/stream object in which DEC file contents can be read
        """
        self.Path = "None:stream_given"
        self.Lines = stream.readlines()
        self._Parse()

    def ParseFile(self, filepath: str) -> None:
        """Parse the supplied file.

        Args:
          filepath (str): path to dec file to parse.  Can be either an absolute path or
            relative to your CWD
        """
        self.Logger.debug("Parsing file: %s" % filepath)
        if not os.path.isabs(filepath):
            fp = self.FindPath(filepath)
        else:
            fp = filepath
        self.Path = fp

        f = open(fp, "r")
        self.Lines = f.readlines()
        f.close()
        self._Parse()



# @file instanced_inf.py
# A module to run a table generator that uses a dsc and environment information to generate a table of information
# about instanced components and libraries where each row is a component or library
##
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""A module to run the InstancedInf table generator against a dsc, adding instanced inf information to the database."""

import logging
import re
from pathlib import Path
from typing import Any

from edk2toollib.database import InstancedInf, Package, Repository, Session, Source
from edk2toollib.database.tables import TableGenerator
from edk2toollib.uefi.edk2.parsers.dsc_parser import DscParser as DscP
from edk2toollib.uefi.edk2.parsers.inf_parser import InfParser as InfP
from edk2toollib.uefi.edk2.path_utilities import Edk2Path


class InstancedInfTable(TableGenerator):
    """A Table Generator that parses a single DSC file and generates a table."""

    SECTION_LIBRARY = "LibraryClasses"
    SECTION_COMPONENT = "Components"
    SECTION_REGEX = re.compile(r"\[(.*)\]")
    OVERRIDE_REGEX = re.compile(r"\<(.*)\>")

    def __init__(self, *args: Any, **kwargs: Any) -> "InstancedInfTable":
        """Initialize the query with the specific settings."""
        self._parsed_infs = {}

    def inf(self, inf: str) -> InfP:
        """Returns a parsed INF object.

        Caches the parsed inf information to reduce multiple re-parses.
        """
        if inf in self._parsed_infs:
            infp = self._parsed_infs[inf]
        else:
            infp = InfP().SetEdk2Path(self.pathobj)
            infp.ParseFile(inf)
            self._parsed_infs[inf] = infp
        return infp

    def parse(self, session: Session, pathobj: Edk2Path, env_id: str, env: dict) -> None:
        """Parse the workspace and update the database."""
        self.pathobj = pathobj
        self.ws = Path(self.pathobj.WorkspacePath)
        self.env = env
        self.dsc = self.env["ACTIVE_PLATFORM"]
        self.arch = self.env["TARGET_ARCH"].split(" ")
        self.target = self.env["TARGET"]

        dscp = DscP().SetEdk2Path(self.pathobj)
        dscp.SetInputVars(self.env)
        dscp.ParseFile(self.dsc)

        # General Debugging
        logging.debug(f"All DSCs included in {self.dsc}:")
        for dsc in dscp.GetAllDscPaths():
            logging.debug(f"  {dsc}")

        logging.debug("Fully expanded DSC:")
        for line in dscp.Lines:
            logging.debug(f"  {line}")
        logging.debug("End of DSC")

        # Parse and insert
        inf_entries = self._build_inf_table(dscp)
        return self._insert_db_rows(session, env_id, inf_entries)

    def _insert_db_rows(self, session: Session, env_id: str, inf_entries: list) -> int:
        """Inserts data into the database.

        Inserts all inf's into the instanced_inf table and links source files and used libraries via the junction
        table.
        """
        # Insert all instanced INF rows
        rows = []
        all_sources = {source.path: source for source in session.query(Source).all()}
        all_packages = {package.name: package for package in session.query(Package).all()}
        all_repos = {
            (repo.name, repo.path): repo for repo in session.query(Repository).filter(Repository.path is not None).all()
        }
        local_repo = session.query(Repository).filter_by(path=None).first()
        for e in inf_entries:
            # Could parse a Windows INF file, which is not a EDKII INF file
            # and won't have a guid. GUIDS are required for INFs so we can
            # assume if it does not have a guid, its the wrong type of INF
            if e["GUID"] == "":
                continue
            sources = []
            for src in e["SOURCES_USED"]:
                if src in all_sources:
                    sources.append(all_sources[src])
                else:
                    sources.append(Source(path=src))
                    all_sources[src] = sources[-1]

            package = None if e["PACKAGE"] is None else all_packages.get(e["PACKAGE"])
            if package is not None:
                repo = package.repository
            else:

                def filter_search(repo: Repository) -> bool:
                    """Return the Repository that contains the INF file."""
                    if repo.path is None:
                        return False
                    return Path(self.pathobj.WorkspacePath, repo.path).as_posix() in Path(e["FULL_PATH"]).as_posix()

                repo = next(
                    filter(filter_search, all_repos.values()),
                    local_repo,  # Default
                )
            rows.append(
                InstancedInf(
                    env=env_id,
                    path=e["PATH"],
                    cls=e.get("LIBRARY_CLASS"),
                    name=e["NAME"],
                    arch=e["ARCH"],
                    dsc=e["DSC"],
                    component=e["COMPONENT"],
                    sources=sources,
                    package=package,
                    repository=repo,
                )
            )
        session.add_all(rows)
        session.commit()

        all_libraries = {
            (library.path, library.arch, library.cls, library.component): library
            for library in session.query(InstancedInf).filter_by(env=env_id).all()
        }
        # Link all instanced INF rows to their used libraries
        for row, inf in zip(rows, inf_entries):
            libraries = [all_libraries.get((path, row.arch, lib, row.component)) for lib, path in inf["LIBRARIES_USED"]]
            row.libraries.extend([lib for lib in libraries if lib is not None])

        session.commit()

    def _build_inf_table(self, dscp: DscP) -> list:
        """Create the instanced inf entries, including components and libraries.

        Multiple entries of the same library will exist if multiple components use it.
        This is where we merge DSC parser information with INF parser information.
        """
        inf_entries = []
        for inf, scope, overrides in dscp.Components:
            # components section scope should only contain the arch.
            # module_type is only needed for libraryclasses section.
            if "." in scope:
                logging.debug(f"DSC section header unnecessarily contains MODULE_TYPE: [Components.{scope.upper()}]")
                scope = scope.split(".")[0]

            # Ignore components built with an architecture that is not in TARGET_ARCH
            arch = scope.upper()
            if arch not in self.arch:
                continue

            # Developers can set an inf path to be relative to any package path. Convert it to be the closest
            # package path relative path to the INF, which is done by `GetEdk2RelativePathFromAbsolutePath`
            inf = self.pathobj.GetAbsolutePathOnThisSystemFromEdk2RelativePath(inf)
            inf = self.pathobj.GetEdk2RelativePathFromAbsolutePath(inf)

            logging.debug(f"Parsing Component: [{arch}][{inf}]")
            infp = InfP().SetEdk2Path(self.pathobj)
            infp.ParseFile(inf)

            # Libraries marked as a component only have source compiled and do not link against other libraries
            if "LIBRARY_CLASS" in infp.Dict:
                continue

            # scope for libraries need to contain the MODULE_TYPE also, so we will append it, if it exists
            if "MODULE_TYPE" in infp.Dict:
                scope += f".{infp.Dict['MODULE_TYPE']}".lower()

            inf_entries += self._parse_inf_recursively(inf, None, inf, dscp.ScopedLibraryDict, overrides, scope, [])

        return inf_entries

    def _parse_inf_recursively(
        self,
        inf: str,
        library_class: str,
        component: str,
        library_dict: dict,
        override_dict: dict,
        scope: str,
        visited: list[str],
    ) -> str:
        """Recurses down all libraries starting from a single INF.

        Will immediately return if the INF has already been visited.
        """
        if inf is None:
            return []

        logging.debug(f"  Parsing Library: [{inf}]")
        visited.append(inf)
        library_instance_list = []
        library_class_list = []
        arch = scope.split(".")[0].upper()

        #
        # 0. Use the existing parser to parse the INF file. This parser parses an INF as an independent file
        #    and does not take into account the context of a DSC.
        #
        infp = self.inf(inf)

        #
        # 1. Convert all libraries to their actual instances for this component. This takes into account
        #    any overrides for this component
        #
        for lib in infp.get_libraries([arch]):
            lib = lib.split(" ")[0]
            library_instance_list.append(self._lib_to_instance(lib.lower(), scope, library_dict, override_dict))
            library_class_list.append(lib)

        #
        # 2. Append all NULL library instances if parsing the component.
        #
        if inf == component:
            for null_lib in override_dict["NULL"]:
                library_instance_list.append(null_lib)
                library_class_list.append("NULL")

            for null_lib in self._get_null_lib_instances(scope, library_dict):
                library_instance_list.append(null_lib)
                library_class_list.append("NULL")

        #
        # 3. Recursively parse used libraries
        #
        to_return = []
        for cls, instance in zip(library_class_list, library_instance_list):
            if instance is None or instance in visited:
                continue
            to_return += self._parse_inf_recursively(
                instance, cls, component, library_dict, override_dict, scope, visited
            )

        # Transform path to edk2 relative form (POSIX)
        def to_posix(path: str) -> str:
            if path is None:
                return None
            return Path(path).as_posix()

        library_instance_list = list(map(to_posix, library_instance_list))

        source_list = []
        full_inf = self.pathobj.GetAbsolutePathOnThisSystemFromEdk2RelativePath(inf)
        pkg = self.pathobj.GetContainingPackage(full_inf)
        for source in infp.get_sources([arch]):
            source = (Path(full_inf).parent / source).resolve()
            source = Path(self.pathobj.GetEdk2RelativePathFromAbsolutePath(str(source))).as_posix()
            source_list.append(source)

        # Return Paths as posix paths, which is Edk2 standard.
        to_return.append(
            {
                "DSC": Path(self.dsc).name,
                "PATH": Path(inf).as_posix(),
                "FULL_PATH": full_inf,
                "GUID": infp.Dict.get("FILE_GUID", ""),
                "NAME": infp.Dict["BASE_NAME"],
                "LIBRARY_CLASS": library_class,
                "COMPONENT": Path(component).as_posix(),
                "MODULE_TYPE": infp.Dict["MODULE_TYPE"],
                "ARCH": arch,
                "PACKAGE": pkg,
                "SOURCES_USED": source_list,
                "LIBRARIES_USED": list(zip(library_class_list, library_instance_list)),
                "PROTOCOLS_USED": [],  # TODO
                "GUIDS_USED": [],  # TODO
                "PPIS_USED": [],  # TODO
                "PCDS_USED": infp.PcdsUsed,
            }
        )
        return to_return

    def _lib_to_instance(self, library_class_name: str, scope: str, library_dict: dict, override_dict: dict) -> str:
        """Converts a library name to the actual instance of the library.

        This conversion is based off the library section definitions in the DSC.
        """
        arch, module = tuple(scope.split("."))

        # NOTE: it is recognized that the below code could be reduced to have less repetitiveness,
        # but I personally believe that the below makes it more clear the order in which we search
        # for matches, and that the order is quite important.

        # https://tianocore-docs.github.io/edk2-DscSpecification/release-1.28/2_dsc_overview/27_[libraryclasses]_section_processing.html#27-libraryclasses-section-processing

        # 1. If a Library class instance (INF) is specified in the Edk2 II [Components] section (an override),
        #    and the library supports the module, then it will be used.
        if library_class_name in override_dict:
            return override_dict[library_class_name]

        # 2/3. If the Library Class instance (INF) is defined in the [LibraryClasses.$(ARCH).$(MODULE_TYPE)] section,
        #      and the library supports the module, then it will be used.
        lookup = f"{arch}.{module}.{library_class_name}"
        if lookup in library_dict:
            library_instance = self._reduce_lib_instances(module, library_dict[lookup])
            if library_instance is not None:
                return library_instance

        # 4. If the Library Class instance (INF) is defined in the [LibraryClasses.common.$(MODULE_TYPE)] section,
        #    and the library supports the module, then it will be used.
        lookup = f"common.{module}.{library_class_name}"
        if lookup in library_dict:
            library_instance = self._reduce_lib_instances(module, library_dict[lookup])
            if library_instance is not None:
                return library_instance

        # 5. If the Library Class instance (INF) is defined in the [LibraryClasses.$(ARCH)] section,
        #    and the library supports the module, then it will be used.
        lookup = f"{arch}.{library_class_name}"
        if lookup in library_dict:
            library_instance = self._reduce_lib_instances(module, library_dict[lookup])
            if library_instance is not None:
                return library_instance

        # 6. If the Library Class Instance (INF) is defined in the [LibraryClasses] section,
        #    and the library supports the module, then it will be used.
        lookup = f"common.{library_class_name}"
        if lookup in library_dict:
            library_instance = self._reduce_lib_instances(module, library_dict[lookup])
            if library_instance is not None:
                return library_instance

        logging.debug(f"scoped library contents: {library_dict}")
        logging.debug(f"override dictionary: {override_dict}")
        e = f"Cannot find library class [{library_class_name}] for scope [{scope}] when evaluating {self.dsc}"
        logging.warning(e)
        return None

    def _reduce_lib_instances(self, module: str, library_instance_list: list[str]) -> str:
        """For a DSC, multiple library instances for the same library class can exist.

        This is either due to a mistake by the developer, or because the library class
        instance only supports certain modules. That is to say a library class instance
        defining `MyLib| PEIM` and one defining `MyLib| PEI_CORE` both being defined in
        the same LibraryClasses section is acceptable.

        Due to this, we need to filter to the first library class instance that supports
        the module type.
        """
        for library_instance in library_instance_list:
            infp = self.inf(library_instance)
            if module.lower() in [phase.lower() for phase in infp.SupportedPhases]:
                # Ensure the returned path is relative to the closest package path
                library_instance = self.pathobj.GetAbsolutePathOnThisSystemFromEdk2RelativePath(library_instance)
                library_instance = self.pathobj.GetEdk2RelativePathFromAbsolutePath(library_instance)
                return library_instance
        return None

    def _get_null_lib_instances(
        self,
        scope: str,
        library_dict: dict,
    ) -> list:
        """Returns all null libraries for a given scope.

        Args:
            scope (str): The scope to search for null libraries.
            library_dict (dict): The dictionary of libraries to search through.

        Returns:
            list: A list of null libraries for the given scope.
        """
        arch, module = tuple(scope.split("."))
        null_libs = []

        lookup = f"{arch}.{module}.null"
        null_libs.extend(library_dict.get(lookup, []))

        lookup = f"common.{module}.null"
        null_libs.extend(library_dict.get(lookup, []))

        lookup = f"{arch}.null"
        null_libs.extend(library_dict.get(lookup, []))

        lookup = "common.null"
        null_libs.extend(library_dict.get(lookup, []))

        return null_libs



# @file variable_format.py
# Module contains helper classes and functions to work with UEFI Variables.
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Module contains helper classes and functions to work with UEFI Variables."""

import struct
import sys
import uuid

import edk2toollib.uefi.uefi_multi_phase as ump

#
# UEFI GUIDs
#
EfiVariableGuid = uuid.UUID(fields=(0xDDCF3616, 0x3275, 0x4164, 0x98, 0xB6, 0xFE85707FFE7D))
EfiAuthenticatedVariableGuid = uuid.UUID(fields=(0xAAF32C78, 0x947B, 0x439A, 0xA1, 0x80, 0x2E144EC37792))

#
# UEFI #Defines
#
HEADER_ALIGNMENT = 4
VARIABLE_STORE_FORMATTED = 0x5A
VARIABLE_STORE_HEALTHY = 0xFE
VARIABLE_DATA = 0x55AA
VAR_IN_DELETED_TRANSITION = 0xFE  # Variable is in obsolete transition.
VAR_DELETED = 0xFD  # Variable is obsolete.
VAR_HEADER_VALID_ONLY = 0x7F  # Variable header has been valid.
VAR_ADDED = 0x3F  # Variable has been completely added.


class VariableStoreHeader(object):
    """Object representing the VARIABLE_STORE_HEADER struct.

    Can parse or produce an VARIABLE_STORE_HEADER structure/byte buffer.

    typedef struct {
        EFI_GUID  Signature;
        UINT32  Size;
        UINT8   Format;
        UINT8   State;
        UINT16  Reserved;
        UINT32  Reserved1;
    } VARIABLE_STORE_HEADER;
    """

    def __init__(self) -> "VariableStoreHeader":
        """Init the empty structure."""
        self.StructString = "=16sLBBHL"  # spell-checker: disable-line
        self.StructSize = struct.calcsize(self.StructString)
        self.Signature = None
        self.Size = None
        self.Format = None
        self.State = None
        self.Reserved0 = None
        self.Reserved1 = None
        self.Type = "Var"

    def load_from_file(self, file: str) -> "VariableStoreHeader":
        """Load the structure from a file."""
        # This function assumes that the file has been seeked
        # to the correct starting location.
        orig_seek = file.tell()
        struct_bytes = file.read(struct.calcsize(self.StructString))
        file.seek(orig_seek)

        # Load this object with the contents of the data.
        (signature_bin, self.Size, self.Format, self.State, self.Reserved0, self.Reserved1) = struct.unpack(
            self.StructString, struct_bytes
        )

        # Update the GUID to be a UUID object.
        if sys.byteorder == "big":
            self.Signature = uuid.UUID(bytes=signature_bin)
        else:
            self.Signature = uuid.UUID(bytes_le=signature_bin)

        # Check one last thing.
        if self.Signature != EfiVariableGuid and self.Signature != EfiAuthenticatedVariableGuid:
            raise Exception("VarStore is of unknown type! %s" % self.Signature)
        if self.Signature == EfiAuthenticatedVariableGuid:
            self.Type = "AuthVar"

        return self

    def serialize(self) -> bytes:
        """Serialize the structure."""
        signature_bin = self.Signature.bytes if sys.byteorder == "big" else self.Signature.bytes_le
        return struct.pack(
            self.StructString, signature_bin, self.Size, self.Format, self.State, self.Reserved0, self.Reserved1
        )


#
# TODO: VariableHeader and AuthenticatedVariableHeader are not truly
#       header structures. They're entire variables. This code should be
#       cleaned up.
#
class VariableHeader(object):
    """Object representing the VARIABLE_HEADER struct.

    Can parse or produce an VARIABLE_HEADER structure/byte buffer.

    typedef struct {
        UINT16      StartId;
        UINT8       State;
        UINT8       Reserved;
        UINT32      Attributes;
        UINT32      NameSize;
        UINT32      DataSize;
        EFI_GUID    VendorGuid;
    } VARIABLE_HEADER;
    """

    def __init__(self) -> "VariableHeader":
        """Init the structure."""
        self.StructString = "=HBBLLL16s"  # spell-checker: disable-line
        self.StructSize = struct.calcsize(self.StructString)
        self.StartId = VARIABLE_DATA
        self.State = VAR_ADDED
        self.Attributes = ump.EFI_VARIABLE_NON_VOLATILE | ump.EFI_VARIABLE_BOOTSERVICE_ACCESS
        self.NameSize = 0
        self.DataSize = 0
        self.VendorGuid = uuid.uuid4()
        self.Name = None
        self.Data = None

    def populate_structure_fields(self, in_bytes: bytes) -> None:
        """Populate the structure field from bytes."""
        (self.StartId, self.State, reserved, self.Attributes, self.NameSize, self.DataSize, self.VendorGuid) = (
            struct.unpack(self.StructString, in_bytes)
        )

    def load_from_bytes(self, in_bytes: bytes) -> "VariableHeader":
        """Load the structure from a bytes."""
        # Load this object with the contents of the data.
        self.populate_structure_fields(in_bytes[0 : self.StructSize])

        # Update the GUID to be a UUID object.
        if sys.byteorder == "big":
            self.VendorGuid = uuid.UUID(bytes=self.VendorGuid)
        else:
            self.VendorGuid = uuid.UUID(bytes_le=self.VendorGuid)

        # Before loading data, make sure that this is a valid variable.
        if self.StartId != VARIABLE_DATA:
            raise EOFError("No variable data!")

        # Finally, load the data.
        data_offset = self.StructSize
        self.Name = in_bytes[data_offset : (data_offset + self.NameSize)].decode("utf-16")
        self.Name = self.Name[:-1]  # Strip the terminating char.
        data_offset += self.NameSize
        self.Data = in_bytes[data_offset : (data_offset + self.DataSize)]

        return self

    def load_from_file(self, file: str) -> "VariableHeader":
        """Load the struct from a file."""
        # This function assumes that the file has been seeked
        # to the correct starting location.
        orig_seek = file.tell()
        struct_bytes = file.read(struct.calcsize(self.StructString))

        # Load this object with the contents of the data.
        self.populate_structure_fields(struct_bytes)

        # Update the GUID to be a UUID object.
        if sys.byteorder == "big":
            self.VendorGuid = uuid.UUID(bytes=self.VendorGuid)
        else:
            self.VendorGuid = uuid.UUID(bytes_le=self.VendorGuid)

        # Before loading data, make sure that this is a valid variable.
        if self.StartId != VARIABLE_DATA:
            file.seek(orig_seek)
            raise EOFError("No variable data!")

        # Finally, load the data.
        self.Name = file.read(self.NameSize).decode("utf-16")[:-1]  # Strip the terminating char.
        self.Data = file.read(self.DataSize)

        file.seek(orig_seek)
        return self

    def get_buffer_data_size(self) -> int:
        """Get the buffer data size."""
        return self.StructSize + self.NameSize + self.DataSize

    def get_buffer_padding_size(self) -> int:
        """Get the buffer padding size."""
        buffer_data_size = self.get_buffer_data_size()
        padding_size = 0
        if buffer_data_size % HEADER_ALIGNMENT != 0:
            padding_size += HEADER_ALIGNMENT - (buffer_data_size % HEADER_ALIGNMENT)
        return padding_size

    def get_buffer_size(self) -> int:
        """Get the buffer size."""
        return self.get_buffer_data_size() + self.get_buffer_padding_size()

    def get_packed_name(self) -> bytes:
        """Get the name attribute."""
        # Make sure to replace the terminating char.
        # name_bytes = b"\x00".join([char for char in (self.Name + b'\x00')])
        name_bytes = self.Name.encode("utf-16")

        # Python encode will leave an "0xFFFE" on the front
        # to declare the encoding type. UEFI does not use this.
        name_bytes = name_bytes[2:]

        # Python encode skips the terminating character, so let's add that.
        name_bytes += b"\x00\x00"

        return name_bytes

    def set_name(self, new_name: str) -> None:
        """Set the name attribute."""
        self.Name = new_name
        self.NameSize = len(self.get_packed_name())

    def set_data(self, new_data: bytes) -> None:
        """Set the data attribute."""
        self.Data = new_data
        self.DataSize = len(new_data)

    def pack_struct(self) -> bytes:
        """Pack the object."""
        vendor_guid = self.VendorGuid.bytes if sys.byteorder == "big" else self.VendorGuid.bytes_le
        return struct.pack(
            self.StructString, self.StartId, self.State, 0, self.Attributes, self.NameSize, self.DataSize, vendor_guid
        )

    def serialize(self, with_padding: bool = False) -> bytes:
        """Serialize the object."""
        bytes = self.pack_struct()

        # Now add the name and data.
        bytes += self.get_packed_name()
        bytes += self.Data

        # Add padding if necessary.
        if with_padding:
            bytes += b"\xff" * self.get_buffer_padding_size()

        return bytes


class AuthenticatedVariableHeader(VariableHeader):
    """Object representing the AUTHENTICATED_VARIABLE_HEADER struct.

    Can parse or produce an AUTHENTICATED_VARIABLE_HEADER structure/byte buffer.

    typedef struct {
        UINT16      StartId;
        UINT8       State;
        UINT8       Reserved;
        UINT32      Attributes;
        UINT64      MonotonicCount;
        EFI_TIME    TimeStamp;
        UINT32      PubKeyIndex;
        UINT32      NameSize;
        UINT32      DataSize;
        EFI_GUID    VendorGuid;
    } AUTHENTICATED_VARIABLE_HEADER;
    """

    def __init__(self) -> "AuthenticatedVariableHeader":
        """Initializes the struct."""
        super(AuthenticatedVariableHeader, self).__init__()
        self.StructString = "=HBBLQ16sLLL16s"  # spell-checker: disable-line
        self.StructSize = struct.calcsize(self.StructString)
        self.MonotonicCount = 0
        self.TimeStamp = b""
        self.PubKeyIndex = 0

    def populate_structure_fields(self, in_bytes: bytes) -> None:
        """Populates the struct."""
        (
            self.StartId,
            self.State,
            reserved,
            self.Attributes,
            self.MonotonicCount,
            self.TimeStamp,
            self.PubKeyIndex,
            self.NameSize,
            self.DataSize,
            self.VendorGuid,
        ) = struct.unpack(self.StructString, in_bytes)

    def pack_struct(self, with_padding: bool = False) -> bytes:
        """Packs the struct."""
        vendor_guid = self.VendorGuid.bytes if sys.byteorder == "big" else self.VendorGuid.bytes_le
        return struct.pack(
            self.StructString,
            self.StartId,
            self.State,
            0,
            self.Attributes,
            self.MonotonicCount,
            self.TimeStamp,
            self.PubKeyIndex,
            self.NameSize,
            self.DataSize,
            vendor_guid,
        )


if __name__ == "__main__":
    pass



# @file BaseParser.py
# Code to support parsing EDK2 files
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Code to support parsing EDK2 files."""

import logging
import os
from typing import Optional, Union
from warnings import warn

from edk2toollib.uefi.edk2 import path_utilities


class BaseParser(object):
    """Base Parser for other parser objects.

    Attributes:
        Parsed (bool): If a file has been parsed or not
        Lines (list): order list of lines in the file
        LocalVars (dict): Dict of local variables
        InputVars (dict): Dict of Input variables
        ConditionalStack (list): list of current condition expressions
        RootPath (str): Workspace root
        PPs (list): List of PPs
        TargetFilePath (list): file being parsed
    """

    operators = ["OR", "AND", "IN", "==", "!=", ">", "<", "<=", ">="]

    def __init__(self, log: str = "BaseParser") -> "BaseParser":
        """Inits an empty Parser."""
        self.Logger = logging.getLogger(log)
        self.Lines = []
        self.LocalVars = {}
        self.InputVars = {}
        self.CurrentSection = ""
        self.CurrentFullSection = ""
        self.Parsed = False
        self.ConditionalStack = []
        self.RootPath = ""
        self.PPs = []
        self._Edk2PathUtil = None
        self.TargetFilePath = None  # the abs path of the target file
        self.CurrentLine = -1
        self._MacroNotDefinedValue = "0"  # value to used for undefined macro

    #
    # For include files set the base root path
    #

    def SetEdk2Path(self, pathobj: path_utilities.Edk2Path) -> "BaseParser":
        """Sets the internal attribute Edk2PathUtil.

        !!! note
            This is a drop in replacement for SetBaseAbsPath and SetPackagePaths as it will asssign both RootPath
            and PPs using the Edk2Path object attributes WorkspacePath and PackagePathList respectively.

        SetBaseAbsPath/SetPackagePaths integration instructions:

        ```python
        # Previous Way
        parser = BaseParser()
        parser.SetBaseAbsPath(path)
        parser.SetPackagePaths(pps)
        ```

        ```python
        # Integration
        parser = BaseParser()
        parser.SetEdk2Path(Edk2Path(path, pps))
        ```

        ```python
        # Integrate with no pps
        parser = BaseParser()
        parser.SetEdk2Path(Edk2Path(path, []))

        Args:
          pathobj (Edk2Path): Edk2Path object

        Returns:
            (BaseParser): self
        """
        self.RootPath = pathobj.WorkspacePath
        self.PPs = pathobj.PackagePathList
        self._Edk2PathUtil = pathobj
        return self

    def SetBaseAbsPath(self, path: str) -> "BaseParser":
        """Sets the attribute RootPath.

        Args:
          path (str): Abs root path

        Returns:
            (BaseParser): self
        """
        warn("SetBaseAbsPath is deprecated.  Use SetEdk2Path instead", DeprecationWarning)
        self.RootPath = os.path.abspath(path)
        self._ConfigEdk2PathUtil()
        return self

    def _ConfigEdk2PathUtil(self) -> None:
        """Creates the path utility object based on the root path and package paths."""
        self._Edk2PathUtil = path_utilities.Edk2Path(self.RootPath, self.PPs, error_on_invalid_pp=False)

    def SetPackagePaths(self, pps: Optional[list[str]] = None) -> "BaseParser":
        """Sets the attribute PPs.

        Args:
          pps (:obj:`list`, optional): list of pps

        NOTE: This must be called after SetBaseAbsPath

        Returns:
            (BaseParser): self
        """
        warn("SetPackagePaths is deprecated.  Use SetEdk2Path instead", DeprecationWarning)
        self.PPs = pps or []
        self._ConfigEdk2PathUtil()
        return self

    def SetInputVars(self, inputdict: dict) -> "BaseParser":
        """Sets the attribute InputVars.

        Args:
          inputdict (dict): The input vars dictionary

        Returns:
            (BaseParser): self
        """
        self.InputVars = inputdict
        return self

    def FindPath(self, *p: str) -> str:
        """Given a path, it will find it relative to the root, the current target file, or the packages path.

        Args:
          *p (obj): any number of strings or path like objects

        Returns:
            (str): a full absolute path if the file exists
            (None): None on failure

        """
        # check if we're getting a None
        if p is None or (len(p) == 1 and p[0] is None):
            return None

        Path = os.path.join(*p)
        # check if it it is absolute
        if os.path.isabs(Path) and os.path.exists(Path):
            return Path

        # If the absolute path exists, return it.
        Path = os.path.join(self.RootPath, *p)
        if os.path.exists(Path):
            return os.path.abspath(Path)

        # If that fails, check a path relative to the target file.
        if self.TargetFilePath is not None:
            Path = os.path.abspath(os.path.join(os.path.dirname(self.TargetFilePath), *p))
            if os.path.exists(Path):
                return os.path.abspath(Path)

        # If that fails, check in every possible Pkg path.
        if self._Edk2PathUtil is not None:
            target_path = os.path.join(*p)
            Path = self._Edk2PathUtil.GetAbsolutePathOnThisSystemFromEdk2RelativePath(target_path, log_errors=False)
            if Path is not None:
                return Path

        # log invalid file path
        self.Logger.error(f"Invalid file path: {p}")
        return None

    def WriteLinesToFile(self, filepath: str) -> None:
        """Write all parsed lines to a file.

        Args:
          filepath (str): path to an unopened file
        """
        self.Logger.debug("Writing all lines to file: %s" % filepath)
        file_handle = open(filepath, "w")
        for line in self.Lines:
            file_handle.write(line + "\n")
        file_handle.close()

    def ComputeResult(self, value: Union[str, int], cond: str, value2: Union[str, int]) -> bool:
        """Compute a logical comaprison.

        Args:
          value: First value
          cond: comparison to do
          value2: Second value

        Returns:
            (bool): result of comparison
        """
        ivalue = value
        ivalue2 = value2
        if isinstance(value, str):
            ivalue = value.strip('"')
        if isinstance(value2, str):
            ivalue2 = value2.strip('"')

        # convert it to interpretted value
        if cond.upper() == "IN":
            # strip quotes
            self.Logger.debug(f"{ivalue} in {ivalue2}")

            return ivalue in ivalue2

        try:
            ivalue = self.ConvertToInt(ivalue)
        except ValueError:
            pass
        try:
            if cond.lower() == "in":
                ivalue2 = set(ivalue2.split())
            else:
                ivalue2 = self.ConvertToInt(ivalue2)
        except ValueError:
            pass

        # First check our boolean operators
        if cond.upper() == "OR":
            return ivalue or ivalue2
        if cond.upper() == "AND":
            return ivalue and ivalue2

        # check our truthyness
        if cond == "==":
            # equal
            return (ivalue == ivalue2) or (value == value2)

        elif cond == "!=":
            # not equal
            return (ivalue != ivalue2) and (value != value2)

        # check to make sure we only have digits from here on out
        if (isinstance(value, str) and value.upper() in ["TRUE", "FALSE"]) or (
            isinstance(value, str) and value2.upper() in ["TRUE", "FALSE"]
        ):
            self.Logger.error(f"Invalid comparison: {value} {cond} {value2}")
            self.Logger.debug(f"Invalid comparison: {value} {cond} {value2}")
            raise ValueError("Invalid comparison")

        if not isinstance(ivalue, int) and not str.isdigit(value):
            self.Logger.error(f"{self.__class__}: Unknown value: {value} {ivalue.__class__}")
            self.Logger.debug(f"{self.__class__}: Conditional: {value} {cond}{value2}")
            raise ValueError("Unknown value")

        if not isinstance(ivalue2, int) and not str.isdigit(value2):
            self.Logger.error(f"{self.__class__}: Unknown value: {value2} {ivalue2}")
            self.Logger.debug(f"{self.__class__}: Conditional: {value} {cond} {value2}")
            raise ValueError("Unknown value")

        if cond == "<":
            return ivalue < ivalue2

        elif cond == "<=":
            return ivalue <= ivalue2

        elif cond == ">":
            return ivalue > ivalue2

        elif cond == ">=":
            return ivalue >= ivalue2

        else:
            self.Logger.error(f"{self.__class__}: Unknown conditional: {cond}")
            raise RuntimeError("Unknown conditional")

    def ConvertToInt(self, value: Union[str, int]) -> int:
        """Converts a str or int to an int based on prefix.

        Args:
          value (str, int): value to convert

        Returns:
            (int): Converted value
        """
        if isinstance(value, int):
            return value
        if isinstance(value, str) and value.upper() == "TRUE":
            return 1
        elif isinstance(value, str) and value.upper() == "FALSE":
            return 0
        elif isinstance(value, str) and value.upper().startswith("0X"):
            return int(value, 16)
        else:
            return int(value, 10)

    def PushConditional(self, v: bool, already_true: bool = False) -> None:
        """Push new value onto the conditional stack.

        Args:
          v (bool): Value to push
          already_true (bool): A boolean that specifies if this condition
            (if/elseif/else) block has already been true.

        !!! note
            already_true is needed when calling PopConditional() to know if the next
            part of the conditional block needs evaluated or not.

        """
        self.ConditionalStack.append((v, already_true))

    def PopConditional(self) -> bool:
        """Pops the current conditional and return the value.

        Additionally returns a value specifying if the if/elseif/else block has already
        returned true.  This is needed to know if the next part of the conditional block
        needs evaluated or not.

        Returns (bool, bool): (value, already_true)
        """
        if len(self.ConditionalStack) > 0:
            return self.ConditionalStack.pop()
        else:
            self.Logger.critical("Tried to pop an empty conditional stack.  Line Number %d" % self.CurrentLine)
            return self.ConditionalStack.pop()  # this should cause a crash but will give trace.

    def _FindReplacementForToken(self, token: str, replace_if_not_found: bool = False) -> str:
        v = self.LocalVars.get(token)

        if v is None:
            v = self.InputVars.get(token)

        if v is None and replace_if_not_found:
            v = self._MacroNotDefinedValue

        elif v is None:
            return None

        if isinstance(v, bool):
            v = "true" if v else "false"

        if isinstance(v, str) and (v.upper() == "TRUE" or v.upper() == "FALSE"):
            v = v.upper()

        return str(v)

    def ReplaceVariables(self, line: str) -> str:
        """Replaces a variable in a string.

        Args:
          line (str): The line to process

        Returns:
            (str): The line with the replaced variable.
        """
        # first tokenize and look for tokens require special macro
        # handling without $.  This must be done first otherwise
        # both syntax options can not be supported.
        result = line
        tokens = result.split()
        replace = len(tokens) > 1 and tokens[0].lower() in ["!ifdef", "!ifndef", "!if", "!elseif"]
        if len(tokens) > 1 and tokens[0].lower() in ["!ifdef", "!ifndef"]:
            if not tokens[1].startswith("$("):
                v = self._FindReplacementForToken(tokens[1], replace)
                if v is not None:
                    result = result.replace(tokens[1], v, 1)

        # use line to avoid change by handling above
        rep = line.count("$")
        index = 0
        while rep > 0:
            start = line.find("$(", index)
            end = line.find(")", start)

            token = line[start + 2 : end]
            replacement_token = line[start : end + 1]
            self.Logger.debug("Token is %s" % token)
            v = self._FindReplacementForToken(token, replace)
            if v is not None:
                result = result.replace(replacement_token, v, 1)

            index = end + 1
            rep = rep - 1

        return result

    def ProcessConditional(self, text: str) -> bool:
        """Processes a conditional.

        Args:
          text (str): The text to process

        Returns:
            (bool): true if a line is a conditiona otherwise false
        """
        if '"' in text:
            tokens = text.split('"')
            tokens = tokens[0].split() + [tokens[1]] + tokens[2].split()
        else:
            tokens = text.split()
        if tokens[0].lower() == "!if":
            result = self.EvaluateConditional(text)
            self.PushConditional(result, result)
            return True

        elif tokens[0].lower() == "!elseif":
            if not self.InActiveCode():
                (_, already_been_true) = self.PopConditional()

                if already_been_true:
                    self.PushConditional(False, True)
                else:
                    result = self.EvaluateConditional(text)
                    self.PushConditional(result, result)
            # already in active code, i.e. the previous if/elseif was true
            else:
                self.PopConditional()
                self.PushConditional(False, True)
            return True

        elif tokens[0].lower() == "!ifdef":
            if len(tokens) != 2:
                self.Logger.error("!ifdef conditionals need to be formatted correctly (spaces between each token)")
                raise RuntimeError("Invalid conditional", text)
            self.PushConditional((tokens[1] != self._MacroNotDefinedValue))
            return True

        elif tokens[0].lower() == "!ifndef":
            if len(tokens) != 2:
                self.Logger.error("!ifdef conditionals need to be formatted correctly (spaces between each token)")
                raise RuntimeError("Invalid conditional", text)
            self.PushConditional((tokens[1] == self._MacroNotDefinedValue))
            return True

        elif tokens[0].lower() == "!else":
            if len(tokens) != 1:
                self.Logger.error("!ifdef conditionals need to be formatted correctly (spaces between each token)")
                raise RuntimeError("Invalid conditional", text)
            (value, already_been_true) = self.PopConditional()
            # TODO make sure we can't do multiple else statements

            # If we've already hit a true condition, we need to be false,
            # otherwise, lets process!
            if already_been_true:
                self.PushConditional(False, True)
            else:
                self.PushConditional(True, True)
            return True

        elif tokens[0].lower() == "!endif":
            if len(tokens) != 1:
                self.Logger.error("!ifdef conditionals need to be formatted correctly (spaces between each token)")
                raise RuntimeError("Invalid conditional", text)
            self.PopConditional()
            return True

        return False

    def EvaluateConditional(self, text: str) -> bool:
        """Uses a pushdown resolver."""
        text = str(text).strip()
        if text.lower().startswith("!if "):
            text = text[3:].strip()
        elif text.lower().startswith("!elseif "):
            text = text[7:].strip()
        else:
            raise RuntimeError(f"Invalid conditional cannot be validated: {text}")

        self.Logger.debug(f"STAGE 1: {text}")
        text = self.ReplaceVariables(text)
        self.Logger.debug(f"STAGE 2: {text}")
        tokens = self._TokenizeConditional(text)
        self.Logger.debug(f"STAGE 3: {tokens}")
        expression = self._ConvertTokensToPostFix(tokens)
        self.Logger.debug(f"STAGE 4: {expression}")

        # Now we evaluate the post fix expression
        if len(expression) == 0:
            raise RuntimeError(f"Malformed !if conditional expression {text} {expression}")
        while len(expression) != 1:
            first_operand_index = -1
            # find the first operator
            for index, item in enumerate(expression):
                if self._IsOperator(item):
                    first_operand_index = index
                    break
            if first_operand_index == -1:
                raise RuntimeError(f"We didn't find an operator to execute in {expression}: {text}")
            operand = expression[first_operand_index]

            if operand == "NOT":
                # Special logic for handling the not
                if first_operand_index < 1:
                    raise RuntimeError(f"We have a stray operand {operand}")
                # grab the operand right before the NOT and invert it
                operator1_raw = expression[first_operand_index - 1]
                operator1 = self.ConvertToInt(operator1_raw)
                result = not operator1
                # grab what was before the operator and the operand, then squish it all together
                new_expression = expression[: first_operand_index - 1] if first_operand_index > 1 else []
                new_expression += [
                    result,
                ] + expression[first_operand_index + 1 :]
                expression = new_expression
            else:
                if first_operand_index < 2:
                    raise RuntimeError(f"We have a stray operand {operand}")
                operator1 = expression[first_operand_index - 2]
                operator2 = expression[first_operand_index - 1]

                do_invert = False
                # check if we have a special operator that has a combined not on it
                if str(operand).startswith("!+"):
                    operand = operand[2:]
                    do_invert = True
                # compute the result now that we have the three things we need
                result = self.ComputeResult(operator1, operand, operator2)

                if do_invert:
                    result = not result
                # grab what was before the operator and the operand, then smoosh it all together
                new_expression = expression[: first_operand_index - 2] if first_operand_index > 2 else []
                new_expression += [
                    result,
                ] + expression[first_operand_index + 1 :]
                expression = new_expression

        final = self.ConvertToInt(expression[0])
        self.Logger.debug(f" FINAL {expression} {final}")

        return bool(final)

    @classmethod
    def _TokenizeConditional(cls: "BaseParser", text: str) -> str:
        """Takes in a string that has macros replaced."""
        # TOKENIZER
        # first we create tokens
        TEXT_MODE = 0
        QUOTE_MODE = 1
        MACRO_MODE = 2
        CONDITION_MODE = 3
        CONDITION_CHARACTERS = ["=", ">", "<", "!"]
        token = ""
        mode = 0
        tokens = []
        for character in text:
            if character == '"' and len(token) == 0 and mode != QUOTE_MODE:
                mode = QUOTE_MODE
            elif character == '"' and mode == QUOTE_MODE:
                if len(token) > 0:
                    tokens.append(f'"{token}"')
                    token = ""
                else:
                    tokens.append('""')
                mode = TEXT_MODE
            elif character == "$" and len(token) == 0:
                token += character
                mode = MACRO_MODE
            elif character == ")" and mode == MACRO_MODE:
                token += character
                tokens.append(token)
                token = ""
                mode = TEXT_MODE
            elif mode == TEXT_MODE and (character == "(" or character == ")"):
                if len(token) > 0:
                    tokens.append(token)
                    token = ""
                tokens.append(character)
            elif character == " " and (mode == TEXT_MODE or mode == MACRO_MODE):
                if len(token) > 0:
                    tokens.append(token)
                    token = ""
                    mode = TEXT_MODE
            elif mode == CONDITION_MODE and character == "(":
                mode = TEXT_MODE
                if len(token) > 0:
                    tokens.append(token)
                    token = ""
                tokens.append(character)
            elif character in CONDITION_CHARACTERS and mode == CONDITION_MODE:
                token += character
            elif character not in CONDITION_CHARACTERS and mode == CONDITION_MODE:
                if len(token) > 0:
                    tokens.append(token)
                token = character if character != " " else ""
                mode = TEXT_MODE
            elif character in CONDITION_CHARACTERS and mode == TEXT_MODE:
                if len(token) > 0:
                    tokens.append(token)
                    token = ""
                token = character
                mode = CONDITION_MODE
            else:
                token += character
        # make sure to add in the last token just in case
        if len(token) > 0:
            tokens.append(token)

        # then we do the lexer and convert operands as necessary
        for index in range(len(tokens)):
            token = tokens[index]
            token_upper = token.upper()
            if token_upper in cls.operators:
                token = token_upper
            elif token_upper == "||":
                token = "OR"
            elif token_upper == "&&":
                token = "AND"
            elif token_upper == "EQ":
                token = "=="
            elif token_upper == "NE":
                token = "!="
            elif token == "!":
                token = "NOT"
            tokens[index] = token

        # collapse the not
        collapsed_tokens = []
        found_not = False
        for token in tokens:
            if str(token).upper() == "NOT":
                found_not = True
                continue
            if not found_not:
                collapsed_tokens.append(token)
            elif token in cls.operators:
                collapsed_tokens.append("!+" + token)
                found_not = False
            else:  # add the not back
                found_not = False
                collapsed_tokens.append("NOT")
                collapsed_tokens.append(token)

        return collapsed_tokens

    @classmethod
    def _ConvertTokensToPostFix(cls: "BaseParser", tokens: list[str]) -> list[str]:
        # convert infix into post fix
        stack = []
        expression = []
        for token in tokens:
            # If the incoming symbol is a left parenthesis, push it on the stack.
            if token == "(":
                stack.append(token)
            # If the incoming symbol is a right parenthesis,
            # pop the stack and print the operators until you see a left parenthesis.
            # Discard the pair of parentheses.
            elif token == ")":
                while len(stack) > 0 and stack[-1] != "(":
                    expression.append(stack.pop())
                stack.pop()  # pop the last (
            # If this isn't a operator ignore it
            elif not cls._IsOperator(token):
                expression.append(token)
            # If the stack is empty or contains a left parenthesis on top, push the incoming operator onto the stack.
            elif len(stack) == 0 or stack[-1] == "(":
                stack.append(token)
            # If the incoming symbol has higher precedence than the top of the stack, push it on the stack.
            elif len(stack) == 0 or cls._GetOperatorPrecedence(token) > cls._GetOperatorPrecedence(stack[-1]):
                stack.append(token)
            # If the incoming symbol has equal precedence with the top of the stack, use association.
            # If the association is left to right, pop and print the top of the stack and
            # then push the incoming operator. If the association is right to left, push the incoming operator.
            elif len(stack) != 0 and cls._GetOperatorPrecedence(token) == cls._GetOperatorPrecedence(stack[-1]):
                expression.append(stack.pop())
                stack.append(token)
            # If the incoming symbol has lower precedence than the symbol on the top of the stack,
            # pop the stack and print the top operator.
            # Then test the incoming operator against the new top of stack.
            elif len(stack) != 0 and cls._GetOperatorPrecedence(token) < cls._GetOperatorPrecedence(stack[-1]):
                while len(stack) > 0 and cls._GetOperatorPrecedence(token) <= cls._GetOperatorPrecedence(stack[-1]):
                    expression.append(stack.pop())
                stack.append(token)
            else:
                logging.error("We don't know what to do with " + token)
        while len(stack) > 0:
            val = stack.pop()
            expression.append(val)
        return expression

    @classmethod
    def _IsOperator(cls: "BaseParser", token: str) -> bool:
        if not isinstance(token, str):
            return False
        if token.startswith("!+"):
            token = token[2:]
        if token == "NOT":  # technically an operator
            return True
        return token in cls.operators

    @classmethod
    def _GetOperatorPrecedence(cls: "BaseParser", token: str) -> int:
        if not cls._IsOperator(token):
            return -1
        if token == "(" or token == ")":
            return 100
        if token == "NOT":  # not is the lowest
            return -2
        if token == "IN" or token == "==" or token == "!=":
            return 1
        return 0

    def InActiveCode(self) -> bool:
        """Determines what the state of the conditional you are currently in.

        Returns:
            (bool): result of the state of the conditional you are in.
        """
        ret = True
        for a, _ in self.ConditionalStack:
            if not a:
                ret = False
                break

        return ret

    def IsGuidString(self, line: str) -> str:
        """Determines if a line is a guid string.

        Args:
            line (str): line representing a possible guid string

        Returns:
            (bool): whether the string is a guid string

        NOTE: format = { 0xD3B36F2C, 0xD551, 0x11D4, { 0x9A, 0x46, 0x00, 0x90, 0x27, 0x3F, 0xC1, 0x4D }}
        Will return true if the the line has
        """
        if line.count("{") == 2 and line.count("}") == 2 and line.count(",") == 10 and line.count("=") == 1:
            return True
        return False

    def ParseGuid(self, line: str) -> str:
        """Parse a guid into a different format.

        Args:
          line (str): the guid to parse ex:
            { 0xD3B36F2C, 0xD551, 0x11D4, { 0x9A, 0x46, 0x00, 0x90, 0x27, 0x3F, 0xC1, 0x4D }}

        Returns:
            (str): guid. ex: D3B36F2C-D551-11D4-9A46-0090273FC14D

        Raises:
            (RuntimeError): if missing any of the 11 parts, or it isn't long enough.
        """
        entries = line.lstrip(" {").rstrip(" }").split(",")
        if len(entries) != 11:
            raise RuntimeError(
                f"Invalid GUID found {line}. We are missing some parts since we only found: {len(entries)}"
            )
        gu = entries[0].lstrip(" 0").lstrip("x").strip()
        # pad front until 8 chars
        while len(gu) < 8:
            gu = "0" + gu

        gut = entries[1].lstrip(" 0").lstrip("x").strip()
        while len(gut) < 4:
            gut = "0" + gut
        gu = gu + "-" + gut

        gut = entries[2].lstrip(" 0").lstrip("x").strip()
        while len(gut) < 4:
            gut = "0" + gut
        gu = gu + "-" + gut

        # strip off extra {
        gut = entries[3].lstrip(" { 0").lstrip("x").strip()
        while len(gut) < 2:
            gut = "0" + gut
        gu = gu + "-" + gut

        gut = entries[4].lstrip(" 0").lstrip("x").strip()
        while len(gut) < 2:
            gut = "0" + gut
        gu = gu + gut

        gut = entries[5].lstrip(" 0").lstrip("x").strip()
        while len(gut) < 2:
            gut = "0" + gut
        gu = gu + "-" + gut

        gut = entries[6].lstrip(" 0").lstrip("x").strip()
        while len(gut) < 2:
            gut = "0" + gut
        gu = gu + gut

        gut = entries[7].lstrip(" 0").lstrip("x").strip()
        while len(gut) < 2:
            gut = "0" + gut
        gu = gu + gut

        gut = entries[8].lstrip(" 0").lstrip("x").strip()
        while len(gut) < 2:
            gut = "0" + gut
        gu = gu + gut

        gut = entries[9].lstrip(" 0").lstrip("x").strip()
        while len(gut) < 2:
            gut = "0" + gut
        gu = gu + gut

        gut = entries[10].split()[0].lstrip(" 0").lstrip("x").rstrip(" } ").strip()
        while len(gut) < 2:
            gut = "0" + gut
        gu = gu + gut

        proper_guid_length = 36
        if len(gu) > proper_guid_length:
            raise RuntimeError(f"The guid we parsed was too long: {gu}")
        if len(gu) < proper_guid_length:
            raise RuntimeError(f"The guid we parsed was too short: {gu}")

        return gu.upper()

    def ResetParserState(self) -> None:
        """Resets the state of the parser."""
        self.ConditionalStack = []
        self.CurrentSection = ""
        self.CurrentFullSection = ""
        self.Parsed = False


class HashFileParser(BaseParser):
    """Base class for Edk2 build files that use # for comments."""

    def __init__(self, log: str) -> "HashFileParser":
        """Inits an empty Parser for files that use # for comments.."""
        BaseParser.__init__(self, log)

    def StripComment(self, line: str) -> str:
        """Removes a comment from a line.

        Args:
          line (str): line with a comment (#)
        """
        if "#" not in line:
            return line.strip()

        result = []
        inside_quotes = False
        quote_char = None
        escaped = False

        for char in line:
            if char in ('"', "'") and not escaped:
                if not inside_quotes:
                    inside_quotes = True
                    quote_char = char
                elif char == quote_char:
                    inside_quotes = False
                    quote_char = None
            elif char == "#" and not inside_quotes:
                break
            elif char == "\\" and not escaped:
                escaped = True
            else:
                escaped = False

            result.append(char)

        return "".join(result).rstrip()

    def ParseNewSection(self, line: str) -> tuple[bool, str]:
        """Parses a new section line.

        Args:
          line (str): line representing a new section.
        """
        if line.count("[") == 1 and line.count("]") == 1:  # new section
            section = line.strip().lstrip("[").split(".")[0].split(",")[0].rstrip("]").strip()
            self.CurrentFullSection = line.strip().lstrip("[").split(",")[0].rstrip("]").strip()
            return (True, section)
        return (False, "")
